from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional

from cryptography import x509
from cryptography.hazmat.primitives import serialization

from uma.cert_utils import get_pubkey
from uma.exceptions import InvalidRequestException
from uma.generated.errors import ErrorCode
from uma.JSONable import JSONable


@dataclass
class PubkeyResponse(JSONable):
    """
    PubkeyResponse is sent from a VASP to another VASP to provide its public keys.
    It is the response to GET requests at `/.well-known/lnurlpubkey`.
    """

    signing_cert_chain: Optional[List[x509.Certificate]]
    """The certificate chain used to verify signatures from a VASP."""

    encryption_cert_chain: Optional[List[x509.Certificate]]
    """The certificate chain used to encrypt TR info sent to a VASP."""

    signing_pubkey: Optional[bytes]
    """Used to verify signatures from a VASP."""

    encryption_pubkey: Optional[bytes]
    """Used to encrypt TR info sent to a VASP."""

    expiration_timestamp: Optional[datetime]
    """
    Optional expiration_timestamp in seconds since epoch at which these pub keys must be refreshed.
	It can be safely cached until this expiration (or forever if null).
    """

    def get_signing_pubkey(self) -> bytes:
        # the first cert in the chain is the leaf (sender's) cert
        if self.signing_cert_chain and self.signing_cert_chain[0]:
            return get_pubkey(self.signing_cert_chain[0])
        if self.signing_pubkey:
            return self.signing_pubkey
        raise InvalidRequestException(
            "Signing pubkey is required for uma.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    def get_encryption_pubkey(self) -> bytes:
        # the first cert in the chain is the leaf (sender's) cert
        if self.encryption_cert_chain and self.encryption_cert_chain[0]:
            return get_pubkey(self.encryption_cert_chain[0])
        if self.encryption_pubkey:
            return self.encryption_pubkey
        raise InvalidRequestException(
            "Encryption pubkey is required for uma.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    def to_dict(self) -> Dict[str, Any]:
        json_dict: Dict[str, Any] = {}
        signing_cert_chain = self.signing_cert_chain
        encryption_cert_chain = self.encryption_cert_chain
        if signing_cert_chain:
            json_dict["signingCertChain"] = [
                cert.public_bytes(encoding=serialization.Encoding.DER).hex()
                for cert in signing_cert_chain
            ]
            json_dict["signingPubKey"] = get_pubkey(signing_cert_chain[0]).hex()
        if encryption_cert_chain:
            json_dict["encryptionCertChain"] = [
                cert.public_bytes(encoding=serialization.Encoding.DER).hex()
                for cert in encryption_cert_chain
            ]
            json_dict["encryptionPubKey"] = get_pubkey(encryption_cert_chain[0]).hex()
        if self.signing_pubkey:
            json_dict["signingPubKey"] = self.signing_pubkey.hex()
        if self.encryption_pubkey:
            json_dict["encryptionPubKey"] = self.encryption_pubkey.hex()
        if self.expiration_timestamp:
            json_dict["expirationTimestamp"] = int(
                self.expiration_timestamp.timestamp()
            )
        return json_dict

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        return {
            "signing_cert_chain": (
                [
                    x509.load_der_x509_certificate(bytes.fromhex(cert))
                    for cert in json_dict["signingCertChain"]
                ]
                if "signingCertChain" in json_dict
                else None
            ),
            "encryption_cert_chain": (
                [
                    x509.load_der_x509_certificate(bytes.fromhex(cert))
                    for cert in json_dict["encryptionCertChain"]
                ]
                if "encryptionCertChain" in json_dict
                else None
            ),
            "signing_pubkey": (
                bytes.fromhex(json_dict["signingPubKey"])
                if "signingPubKey" in json_dict
                else None
            ),
            "encryption_pubkey": (
                bytes.fromhex(json_dict["encryptionPubKey"])
                if "encryptionPubKey" in json_dict
                else None
            ),
            "expiration_timestamp": (
                datetime.fromtimestamp(json_dict["expirationTimestamp"], timezone.utc)
                if "expirationTimestamp" in json_dict
                else None
            ),
        }



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Dict, Optional

from uma.protocol.pubkey_response import PubkeyResponse


class IPublicKeyCache(ABC):
    @abstractmethod
    def fetch_public_key_for_vasp(self, vasp_domain: str) -> Optional[PubkeyResponse]:
        pass

    @abstractmethod
    def add_public_key_for_vasp(
        self, vasp_domain: str, public_key: PubkeyResponse
    ) -> None:
        pass

    @abstractmethod
    def remove_public_key_for_vasp(self, vasp_domain: str) -> None:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass


class InMemoryPublicKeyCache(IPublicKeyCache):
    def __init__(self) -> None:
        self._cache: Dict[str, PubkeyResponse] = {}

    def fetch_public_key_for_vasp(self, vasp_domain: str) -> Optional[PubkeyResponse]:
        public_key = self._cache.get(vasp_domain, None)
        if not public_key:
            return None

        return (
            public_key
            if (
                not public_key.expiration_timestamp
                or public_key.expiration_timestamp > datetime.now(timezone.utc)
            )
            else None
        )

    def add_public_key_for_vasp(
        self, vasp_domain: str, public_key: PubkeyResponse
    ) -> None:
        self._cache[vasp_domain] = public_key

    def remove_public_key_for_vasp(self, vasp_domain: str) -> None:
        if vasp_domain in self._cache:
            self._cache.pop(vasp_domain)

    def clear(self) -> None:
        self._cache.clear()


class IAsyncPublicKeyCache(ABC):
    @abstractmethod
    async def gen_fetch_public_key_for_vasp(
        self, vasp_domain: str
    ) -> Optional[PubkeyResponse]:
        pass

    @abstractmethod
    async def gen_add_public_key_for_vasp(
        self, vasp_domain: str, public_key: PubkeyResponse
    ) -> None:
        pass

    @abstractmethod
    async def gen_remove_public_key_for_vasp(self, vasp_domain: str) -> None:
        pass

    @abstractmethod
    async def gen_clear(self) -> None:
        pass



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from dataclasses import dataclass
from typing import Dict
from enum import Enum

from uma.JSONable import JSONable


@dataclass
class CounterpartyDataOption(JSONable):
    mandatory: bool
    """Whether the field is mandatory or not"""


CounterpartyDataOptions = Dict[str, CounterpartyDataOption]


class CounterpartyDataKeys(Enum):
    """Common keys used in counterparty data exchanges between VASPs."""

    IDENTIFIER = "identifier"
    """The UMA address of the counterparty"""

    NAME = "name"
    """The full name of the counterparty"""

    EMAIL = "email"
    """The email address of the counterparty"""

    COMPLIANCE = "compliance"
    """Compliance-related data including KYC status, UTXOs, and travel rule information"""

    BIRTH_DATE = "birthDate"
    """The counterparty's date of birth, in ISO 8601 format"""

    NATIONALITY = "nationality"
    """The counterparty's nationality, in ISO 3166-1 alpha-2 format"""

    COUNTRY_OF_RESIDENCE = "countryOfResidence"
    """The counterparty's country of residence, in ISO 3166-1 alpha-2 format"""

    PHONE_NUMBER = "phoneNumber"
    """The counterparty's phone number, in E.164 format"""

    FI_LEGAL_ENTITY_NAME = "fiLegalEntityName"
    """The counterparty financial institution's legal entity name"""


def create_counterparty_data_options(
    options: Dict[str, bool],
) -> CounterpartyDataOptions:
    return {
        key: CounterpartyDataOption(mandatory=value) for key, value in options.items()
    }



from dataclasses import dataclass
from typing import List, Optional
from enum import Enum

from uma.JSONable import JSONable


class TransactionStatus(str, Enum):
    COMPLETED = "COMPLETED"
    """
    Recipient received the payment.
    """

    FAILED = "FAILED"
    """
    Payment failed due to a post-transaction error.
    """


@dataclass
class UtxoWithAmount(JSONable):
    utxo: str
    amount_msats: int


@dataclass
class PostTransactionCallback(JSONable):
    utxos: List[UtxoWithAmount]
    """List of utxo/amounts corresponding to the VASPs channels."""

    vasp_domain: Optional[str]
    """
    The domain of the VASP that is sending the callback. Used by the VASP to fetch the public keys of
    its counterparty.

    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    signature: Optional[str]
    """
    Signature is the base64-encoded signature of sha256(Nonce|Timestamp)

    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    signature_nonce: Optional[str]
    """
    Random string that is used to prevent replay attacks

    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    signature_timestamp: Optional[int]
    """
    Timestamp of the signature in seconds since epoch
    
    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    transaction_status: Optional[TransactionStatus]
    """
    The status of the transaction.
    """

    error_code: Optional[str]
    """
    The error code if the transaction has failed. This should be one of the [ErrorCode] enum values.
    This should only be set if `transaction_status` is `FAILED`.
    """

    error_reason: Optional[str]
    """
    The reason for the transaction failure.
    Only set when transaction_status is `FAILED`.
    """

    def signable_payload(self) -> bytes:
        if (
            not self.signature_nonce
            or not self.signature_timestamp
            or not self.vasp_domain
        ):
            raise ValueError(
                "Fields needed for signature are missing. Nonce, timestamp, and vasp_domain are required for UMA v1."
            )
        signable = "|".join([self.signature_nonce, str(self.signature_timestamp)])
        return signable.encode("utf-8")



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from dataclasses import dataclass

from uma.JSONable import JSONable


@dataclass
class SettlementInfo(JSONable):
    layer: str
    """
    The settlement layer chosen by the sender (e.g., "ln", "spark").
    """

    asset_identifier: str
    """
    The identifier of the settlement asset chosen by the sender.
    """

    @classmethod
    def _get_field_name_overrides(cls) -> dict[str, str]:
        return {
            "asset_identifier": "assetIdentifier",
        }


@dataclass
class SettlementAsset(JSONable):
    identifier: str
    """
    The identifier of the asset. For Lightning, this should be "BTC".
    For Spark, this is the token identifier.
    """

    multipliers: dict[str, float]
    """
    Estimated conversion rates from this asset to the currencies supported by
    the receiver. The key is the currency code and the value is the multiplier
    (how many of the smallest unit of this asset equals one unit of the
    currency).
    """


@dataclass
class SettlementOption(JSONable):
    settlement_layer: str
    """
    The name of the settlement layer (e.g., "spark", "ln").
    """

    assets: list[SettlementAsset]
    """
    List of accepted assets on this settlement layer with their conversion rates.
    """

    @classmethod
    def _get_field_name_overrides(cls) -> dict[str, str]:
        return {
            "settlement_layer": "settlementLayer",
        }



from abc import ABC
from typing import (
    Optional,
    Protocol,
    Any,
    Dict,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    runtime_checkable,
)
import struct

T = TypeVar("T", bound="TLVCodable")


@runtime_checkable
class ByteCodable(Protocol):
    def to_bytes(self) -> bytes: ...

    @classmethod
    def from_bytes(cls, data: bytes) -> "ByteCodable": ...


class TLVCodable(ABC):
    def to_tlv(self) -> bytes:
        result = bytearray()
        tlv_map = self.tlv_map()
        for attr_name, _ in self.__dict__.items():
            value = getattr(self, attr_name)
            if value is not None:
                tag = tlv_map[attr_name]
                encoded_value = self._encode_value(value)
                length = len(encoded_value)
                result.extend(struct.pack("!BB", tag, length))
                result.extend(encoded_value)
        return bytes(result)

    @classmethod
    def from_tlv(cls: "type[T]", data: bytes) -> T:
        obj = cls()
        index = 0
        tag_to_data: Dict[int, bytes] = {}
        while index + 2 < len(data):
            tag, length = struct.unpack("!BB", data[index : index + 2])
            index += 2
            value_bytes = data[index : index + length]
            index += length
            tag_to_data[tag] = value_bytes

        tlv_map = cls.tlv_map()
        for attr_name, tag in tlv_map.items():
            value_bytes = tag_to_data.get(tag, None)
            if value_bytes is None:
                continue
            attr_type = cls._get_attribute_type(obj, attr_name)
            value = cls._decode_value(value_bytes, attr_type)
            setattr(obj, attr_name, value)

        return obj

    @classmethod
    def _get_attribute_type(cls, obj, attribute_name):
        # Get type hints for the object
        type_hints = get_type_hints(type(obj))

        # Check if the attribute exists in type hints
        if attribute_name in type_hints:
            attr_type = type_hints[attribute_name]

            # Check if it's an Optional type
            if hasattr(attr_type, "__origin__") and attr_type.__origin__ is Optional:
                return attr_type.__args__[0]
            return attr_type
        # Fallback to runtime type checking
        return type(getattr(obj, attribute_name))

    @classmethod
    def tlv_map(cls) -> Dict[str, int]:
        return {}

    @classmethod
    def _encode_value(cls, value: Any) -> bytes:
        if isinstance(value, int):
            if -128 <= value <= 127:
                result = struct.pack("!b", value)  # int8
            elif -32768 <= value <= 32767:
                result = struct.pack("!h", value)  # int16
            elif -2147483648 <= value <= 2147483647:
                result = struct.pack("!i", value)  # int32
            else:
                result = struct.pack("!q", value)  # int64
        elif isinstance(value, float):
            result = struct.pack("!d", value)  # 8-byte float
        elif isinstance(value, bytes):
            result = value
        elif isinstance(value, bool):
            result = struct.pack("!?", value)
        elif isinstance(value, str):
            result = value.encode("utf-8")
        elif isinstance(value, TLVCodable):
            result = value.to_tlv()
        elif isinstance(value, ByteCodable):
            result = value.to_bytes()
        else:
            raise ValueError(f"Unsupported type: {type(value)}")
        return result

    @classmethod
    def _decode_value(cls, value: bytes, attr_type: Type) -> Any:
        attr_type = unwrap_type(attr_type)
        if attr_type == int:
            result = int.from_bytes(value, byteorder="big")
        elif attr_type == float:
            result = struct.unpack("!d", value)[0]
        elif attr_type == str:
            result = value.decode("utf-8")
        elif attr_type == bytes:
            result = value
        elif attr_type == bool:
            result = struct.unpack("!?", value)[0]
        elif issubclass(attr_type, TLVCodable):
            result = attr_type.from_tlv(value)
        elif issubclass(attr_type, ByteCodable):
            result = attr_type.from_bytes(value)
        else:
            raise ValueError(f"Unsupported type: {attr_type}")
        return result


def unwrap_type(typ):
    origin = get_origin(typ)
    if origin is Union or origin is Optional:
        args = get_args(typ)
        return unwrap_type(args[0])  # Recursively unwrap
    return typ



import json
from dataclasses import dataclass
from typing import Any, Dict, Optional

from uma.exceptions import InvalidRequestException
from uma.generated.errors import ErrorCode
from uma.JSONable import JSONable
from uma.protocol.backing_signature import BackingSignature
from uma.protocol.counterparty_data import CounterpartyDataOptions, CounterpartyDataKeys
from uma.protocol.payer_data import PayerData, compliance_from_payer_data
from uma.protocol.settlement import SettlementInfo
from uma.protocol.v0.payreq import PayRequest as V0PayRequest
from uma.protocol.v1.payreq import PayRequest as V1PayRequest
from uma.signing_utils import sign_payload
from uma.type_utils import none_throws
from uma.version import MAJOR_VERSION


@dataclass
class PayRequest(JSONable):
    sending_amount_currency_code: Optional[str]
    """
    The currency code of the `amount` field. `None` indicates that `amount` is in the smallest
    unit of the settlement asset. For lightning, this is millisatoshis as in LNURL without LUD-21.
    If this is not `None`, then `amount` is in the smallest unit of
    the specified currency (e.g. cents for USD). This currency code can be any currency which
    the receiver can quote. However, there are two most common scenarios for UMA:

    1. If the sender wants the receiver wants to receive a specific amount in their receiving
    currency, then this field should be the same as `receiving_currency_code`. This is useful
    for cases where the sender wants to ensure that the receiver receives a specific amount
    in that destination currency, regardless of the exchange rate, for example, when paying
    for some goods or services in a foreign currency.

    2. If the sender has a specific amount in their own currency that they would like to send,
    then this field should be left as `None` to indicate that the amount is in the smallest
    unit of the settlement asset (ie. msats by default).
    This will lock the sent amount on the sender side, and the receiver will receive the
    equivalent amount in their receiving currency. NOTE: In this scenario, the sending VASP
    *should not* pass the sending currency code here, as it is not relevant to the receiver.
    Rather, by specifying an invoice amount in the settlement asset (for example, msats for
    lightning), the sending VASP can ensure that their user will be sending a fixed amount,
    regardless of the exchange rate on the receiving side.
    """

    receiving_currency_code: Optional[str]
    """
    The currency code for the currency that the receiver will receive for this payment.
    """

    amount: int
    """
    The amount of the payment in the currency specified by `currency_code`. This amount is
    in the smallest unit of the specified currency (e.g. cents for USD).
    """

    payer_data: Optional[PayerData]
    """
    The data about the payer that the sending VASP must provide in order to send a payment.
    This was requested by the receiver in the lnulp response. See LUD-18.
    """

    requested_payee_data: Optional[CounterpartyDataOptions]
    """
    The data about the receiver that the sending VASP would like to know from the receiver.
    See LUD-22.
    """

    comment: Optional[str] = None
    """
    A comment that the sender would like to include with the payment. This can only be included
    if the receiver included the `commentAllowed` field in the lnurlp response. The length of
    the comment must be less than or equal to the value of `commentAllowed`.
    """

    uma_major_version: Optional[int] = MAJOR_VERSION
    """
    The major version of the UMA protocol that this currency adheres to. This is not serialized to JSON.
    """

    invoice_uuid: Optional[str] = None
    """
    The uma invoice UUID that the sender is paying.
    """

    settlement_info: Optional[SettlementInfo] = None
    """
    Settlement information including the layer and asset chosen by the sender.
    Must be one of the options provided by the receiver in the lnurlp response.
    If not specified, defaults to Lightning with BTC.
    """

    def signable_payload(self) -> bytes:
        if not self.payer_data:
            raise InvalidRequestException(
                "payer_data is required.", ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS
            )
        payer_identifier = self.payer_data.get(CounterpartyDataKeys.IDENTIFIER.value)
        if not payer_identifier:
            raise InvalidRequestException(
                "identifier is required in payerdata for uma.",
                ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
            )
        payloads = [payer_identifier]
        compliance = compliance_from_payer_data(none_throws(self.payer_data))
        if compliance:
            payloads += [
                compliance.signature_nonce,
                str(compliance.signature_timestamp),
            ]
        return "|".join(payloads).encode("utf8")

    def is_uma_request(self) -> bool:
        return (
            self.payer_data is not None
            and CounterpartyDataKeys.COMPLIANCE.value in self.payer_data
        )

    def to_dict(self) -> Dict[str, Any]:
        version_payreq = (
            V0PayRequest(
                currency_code=self.receiving_currency_code,
                amount=self.amount,
                payer_data=self.payer_data,
                requested_payee_data=self.requested_payee_data,
                comment=self.comment,
            )
            if self.uma_major_version == 0
            else V1PayRequest(
                sending_amount_currency_code=self.sending_amount_currency_code,
                receiving_currency_code=self.receiving_currency_code,
                amount=self.amount,
                payer_data=self.payer_data,
                requested_payee_data=self.requested_payee_data,
                comment=self.comment,
                settlement_info=self.settlement_info,
            )
        )
        return version_payreq.to_dict()

    @classmethod
    def from_json(cls: "type[PayRequest]", json_encoded: str) -> "PayRequest":
        json_dict = json.loads(json_encoded)
        is_amount_string = "amount" in json_dict and isinstance(
            json_dict["amount"], str
        )
        is_uma = json_dict.get("payerData") and "compliance" in json_dict["payerData"]
        is_v1 = "convert" in json_dict and is_uma
        is_v0 = "currency" in json_dict and is_uma

        if is_v1 or is_amount_string:
            v1_payreq = V1PayRequest.from_json(json_encoded)
            return PayRequest(
                sending_amount_currency_code=v1_payreq.sending_amount_currency_code,
                receiving_currency_code=v1_payreq.receiving_currency_code,
                amount=v1_payreq.amount,
                payer_data=v1_payreq.payer_data,
                requested_payee_data=v1_payreq.requested_payee_data,
                comment=v1_payreq.comment,
                settlement_info=v1_payreq.settlement_info,
                uma_major_version=1 if is_v1 else None,
            )
        v0_payreq = V0PayRequest.from_json(json_encoded)
        return PayRequest(
            sending_amount_currency_code=v0_payreq.currency_code,
            receiving_currency_code=v0_payreq.currency_code,
            amount=v0_payreq.amount,
            payer_data=v0_payreq.payer_data,
            requested_payee_data=None,
            uma_major_version=0 if is_v0 else None,
        )

    def to_request_params(self) -> Dict[str, str]:
        params = {}
        if self.sending_amount_currency_code:
            amount = f"{self.amount}.{self.sending_amount_currency_code}"
        else:
            amount = str(self.amount)
        params["amount"] = amount
        if self.receiving_currency_code:
            params["convert"] = self.receiving_currency_code
        if self.payer_data is not None:
            params["payerData"] = json.dumps(self.payer_data)
        if self.requested_payee_data is not None:
            params["payeeData"] = json.dumps(self.requested_payee_data)
        if self.comment:
            params["comment"] = self.comment
        return params

    @classmethod
    def from_request_params(
        cls: "type[PayRequest]", params: Dict[str, str]
    ) -> "PayRequest":
        if not params.get("amount"):
            raise InvalidRequestException(
                "amount is required.", ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS
            )
        parts = params["amount"].split(".")
        sending_amount_currency_code = parts[1] if len(parts) == 2 else None
        amount = parts[0]
        payer_data_json = params.get("payerData")
        payer_data = None if payer_data_json is None else json.loads(payer_data_json)
        payee_data_json = params.get("payeeData")
        payee_data = None if payee_data_json is None else json.loads(payee_data_json)
        return PayRequest(
            sending_amount_currency_code=sending_amount_currency_code,
            receiving_currency_code=params.get("convert"),
            amount=int(amount),
            payer_data=payer_data,
            comment=params.get("comment"),
            requested_payee_data=payee_data,
        )

    def append_backing_signature(self, signing_private_key: bytes, domain: str) -> None:
        """
        Appends a backing signature to the payreq request.

        Args:
            signing_private_key: The private key of the backing VASP which is used to sign the payload.
            domain: The domain of the backing VASP that produced the signature. Public keys for this VASP
            will be fetched from this domain at /.well-known/lnurlpubkey and used to verify the signature.
        """
        if not self.is_uma_request():
            return
        payer_data = self.payer_data
        if not payer_data:
            raise InvalidRequestException(
                "payer_data is required.", ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS
            )
        compliance = compliance_from_payer_data(payer_data)
        if not compliance:
            raise InvalidRequestException(
                "compliance data is required for backing signatures",
                ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
            )
        payload = self.signable_payload()
        backing_signature = sign_payload(payload, signing_private_key)
        if compliance.backing_signatures is None:
            compliance.backing_signatures = []
        compliance.backing_signatures.append(
            BackingSignature(domain=domain, signature=backing_signature)
        )
        payer_data[CounterpartyDataKeys.COMPLIANCE.value] = compliance.to_dict()



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from dataclasses import dataclass
import json
from typing import Any, Dict, List, Optional

from uma.JSONable import JSONable
from uma.protocol.backing_signature import BackingSignature
from uma.protocol.counterparty_data import CounterpartyDataKeys
from uma.protocol.kyc_status import KycStatus


@dataclass
class CompliancePayerData(JSONable):
    kyc_status: KycStatus
    """KYC information about the sender."""

    utxos: List[str]
    """The list of UTXOs of the sender's channels that might be used to forward the payment"""

    node_pubkey: Optional[str]
    """The public key of the sender's node that will be used to send the payment"""

    encrypted_travel_rule_info: Optional[str]
    """The travel rule information of the sender. This is encrypted with the receiver's public encryption key"""

    travel_rule_format: Optional[str]
    """
    An optional standardized format of the travel rule information (e.g. IVMS). Null indicates raw json or a custom format.
    This field is formatted as <standardized format>@<version> (e.g. ivms@101.2023). Version is optional.
    """

    signature: str
    """Signature is the base64-encoded signature of sha256(ReceiverAddress|Nonce|Timestamp)"""

    signature_nonce: str

    signature_timestamp: int
    """Time stamp of the signature in seconds"""

    utxo_callback: str
    """The URL that the receiver will call to send UTXOs of the channels that receiver used to receive the payment the once it completes."""

    backing_signatures: Optional[List[BackingSignature]] = None
    """List of backing VASP signatures."""

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {"node_pubkey": "nodePubKey"}

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        data = super()._from_dict(json_dict)
        # Some SDKs allow empty lists to be omitted from the JSON, so we need to add them back in.
        # This will be reconciled in the next major version bump.
        if "utxos" not in data or not data["utxos"]:
            data["utxos"] = []
        return data


PayerData = Dict[str, Any]


def compliance_from_payer_data(payer_data: PayerData) -> Optional[CompliancePayerData]:
    compliance = payer_data.get(CounterpartyDataKeys.COMPLIANCE.value)
    if not compliance or not isinstance(compliance, dict):
        return None

    return CompliancePayerData.from_json(json.dumps(compliance))


def create_payer_data(
    identifier: str,
    email: Optional[str] = None,
    name: Optional[str] = None,
    compliance: Optional[CompliancePayerData] = None,
) -> PayerData:
    payer_data: PayerData = {
        CounterpartyDataKeys.IDENTIFIER.value: identifier,
    }
    if email:
        payer_data[CounterpartyDataKeys.EMAIL.value] = email
    if name:
        payer_data[CounterpartyDataKeys.NAME.value] = name
    if compliance:
        payer_data[CounterpartyDataKeys.COMPLIANCE.value] = compliance.to_dict()
    return payer_data



from dataclasses import dataclass
import json
from typing import Any, Dict, List, Optional

from uma.JSONable import JSONable
from uma.exceptions import InvalidRequestException
from uma.generated.errors import ErrorCode
from uma.protocol.backing_signature import BackingSignature
from uma.protocol.payee_data import PayeeData
from uma.protocol.counterparty_data import CounterpartyDataKeys
from uma.signing_utils import sign_payload
from uma.version import MAJOR_VERSION


@dataclass
class PayReqResponseCompliance(JSONable):
    utxos: List[str]
    """List of UTXOs of channels over which the receiver will likely receive the payment"""

    utxo_callback: str
    """URL that the sender VASP will call to send UTXOs of the channel that the sender used to send the payment once it completes"""

    node_pubkey: Optional[str]
    """Public key of the receiver's node if known"""

    signature: Optional[str]
    """
    Signature is the base64-encoded signature of sha256(ReceiverAddress|Nonce|Timestamp)

    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    signature_nonce: Optional[str]
    """
    Random string that is used to prevent replay attacks

    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    signature_timestamp: Optional[int]
    """
    Time stamp of the signature in seconds since epoch

    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    backing_signatures: Optional[List[BackingSignature]] = None
    """
    List of backing VASP signatures.
    """

    def signable_payload(self, sender_address: str, receiver_address: str) -> bytes:
        if not self.signature_nonce or not self.signature_timestamp:
            raise ValueError(
                "Compliance data is missing signature nonce or timestamp fields. These are required for UMA v1."
            )
        signable = "|".join(
            [
                sender_address,
                receiver_address,
                self.signature_nonce,
                str(int(self.signature_timestamp)),
            ]
        )
        return signable.encode("utf8")

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {"node_pubkey": "nodePubKey"}

    @classmethod
    def from_payee_data(
        cls, payee_data: PayeeData
    ) -> Optional["PayReqResponseCompliance"]:
        compliance = payee_data.get(CounterpartyDataKeys.COMPLIANCE.value)
        if not compliance or not isinstance(compliance, dict):
            return None
        return PayReqResponseCompliance.from_json(json.dumps(compliance))


@dataclass
class PayReqResponsePaymentInfo(JSONable):
    amount: Optional[int]
    """
    The amount that the receiver will receive in the receiving currency not including fees. The amount is specified
    in the smallest unit of the currency (eg. cents for USD).

    Note: This field is optional for UMA v0.X backwards-compatibility. It is required for UMA v1.X.
    """

    currency_code: str
    """
    The currency code that the receiver will receive for this payment.
    """

    decimals: int
    """
    Number of digits after the decimal point for the receiving currency. For example, in USD, by
    convention, there are 2 digits for cents - $5.95. In this case, `decimals` would be 2. This should align with
    the currency's `decimals` field in the LNURLP response. It is included here for convenience. See
    [UMAD-04](https://github.com/uma-universal-money-address/protocol/blob/main/umad-04-lnurlp-response.md) for
    details, edge cases, and examples.
    """

    multiplier: float
    """
    The conversion rate. In the default case (Lightning/BTC), it is the number of millisatoshis that the
    receiver will receive for 1 unit of the specified currency (eg: cents in USD). In this context, this
    is just for convenience. The conversionrate is also baked into the invoice amount itself. Specifically:
    `invoiceAmount = amount * multiplier + exchange_fees`

    For other settlement layers, this is the number of the smallest unit of the settlement asset that the
    receiver will receive for 1 unit of the specified currency.
    """

    exchange_fees_msats: int
    """
    The fees charged by the receiving VASP for this transaction.
    Deprecated: Use `exchange_fees` instead. This field will be removed in the next major version.
    """

    @property
    def exchange_fees(self) -> int:
        """
        The fees charged by the receiving VASP for this transaction. In the default case (Lightning/BTC),
        this is in millisatoshis. For other settlement layers, this is in the smallest unit of the
        settlement asset. This is separate from the `multiplier`.
        """
        return self.exchange_fees_msats

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {"exchange_fees_msats": "fee"}

    def to_dict(self) -> Dict[str, Any]:
        resp = super().to_dict()
        # For backwards-compatibility with UMA v0, duplicate the fee field
        # under both names:
        resp["exchangeFeesMillisatoshi"] = self.exchange_fees
        return resp

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        # For backwards-compatibility with UMA v0, pull fees from v0 name if
        # not present in v1 name:
        if "fee" not in json_dict:
            json_dict["fee"] = json_dict.pop("exchangeFeesMillisatoshi", 0)

        return super()._from_dict(json_dict)


@dataclass
class PayReqResponse(JSONable):
    encoded_invoice: str
    """
    The encoded BOLT11 invoice that the sender will use to pay the receiver.
    """

    routes: List[str]
    """
    Always just an empty array for legacy reasons.
    """

    payee_data: Optional[PayeeData]
    """
    The data about the receiver that the sending VASP requested in the payreq request.
    Required for UMA.
    """

    payment_info: Optional[PayReqResponsePaymentInfo]
    """
    Information about the payment that the receiver will receive. Includes
    Final currency-related information for the payment. Required for UMA.
    """

    disposable: Optional[bool] = None
    """
    This field may be used by a WALLET to decide whether the initial LNURL link will
    be stored locally for later reuse or erased. If disposable is null, it should be
    interpreted as true, so if SERVICE intends its LNURL links to be stored it must
    return `disposable: false`. UMA should never return `disposable: false` due to
    signature nonce checks, etc. See LUD-11.
    """

    success_action: Optional[Dict[str, str]] = None
    """
    Defines a struct which can be stored and shown to the user on payment success. See LUD-09.
    """

    uma_major_version: Optional[int] = MAJOR_VERSION
    """
    The major version of the UMA protocol that this currency adheres to. This is not serialized to JSON.
    """

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {"encoded_invoice": "pr", "payment_info": "converted"}

    def is_uma_response(self) -> bool:
        return (
            self.payee_data is not None
            and CounterpartyDataKeys.COMPLIANCE.value in self.payee_data
            and self.payment_info is not None
        )

    def get_compliance(self) -> Optional[PayReqResponseCompliance]:
        if not self.payee_data:
            return None
        return PayReqResponseCompliance.from_payee_data(self.payee_data)

    def append_backing_signature(
        self,
        signing_private_key: bytes,
        domain: str,
        sender_address: str,
        receiver_address: str,
    ) -> None:
        """
        Appends a backing signature to the payreq response.

        Args:
            signing_private_key: The private key of the backing VASP which is used to sign the payload.
            domain: The domain of the backing VASP that produced the signature. Public keys for this VASP
            will be fetched from this domain at /.well-known/lnurlpubkey and used to verify the signature.
            sender_address: The sender's UMA address.
            receiver_address: The receiver's UMA address.
        """
        if not self.is_uma_response():
            return
        compliance = self.get_compliance()
        if compliance is None:
            raise InvalidRequestException(
                "compliance data is required for adding backing signatures",
                ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
            )
        backing_signatures = compliance.backing_signatures
        if backing_signatures is None:
            backing_signatures = []
        payload = compliance.signable_payload(sender_address, receiver_address)
        backing_signature = sign_payload(payload, signing_private_key)
        backing_signatures.append(
            BackingSignature(domain=domain, signature=backing_signature)
        )
        compliance.backing_signatures = backing_signatures
        payee_data = self.payee_data or {}
        payee_data[CounterpartyDataKeys.COMPLIANCE.value] = compliance.to_dict()
        self.payee_data = payee_data

    def to_dict(self) -> Dict[str, Any]:
        resp = super().to_dict()
        # For backwards-compatibility with UMA v0, move the compliance data
        # and payment info to v0 names:
        if self.uma_major_version == 0:
            compliance = self.get_compliance()
            if compliance:
                resp["compliance"] = compliance.to_dict()
                resp["payeeData"].pop("compliance", None)
            if self.payment_info:
                resp["paymentInfo"] = self.payment_info.to_dict()
                resp.pop("converted", None)
        resp.pop("umaMajorVersion", None)

        return resp

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        data = super()._from_dict(json_dict)

        # For backwards-compatibility with UMA v0, duplicate the compliance data
        # from the top-level response over to payee_data:
        is_uma_v0 = "compliance" in json_dict
        is_uma_v1 = "payeeData" in json_dict and "compliance" in json_dict["payeeData"]
        if is_uma_v0 and not is_uma_v1:
            if "payee_data" not in data or not data["payee_data"]:
                data["payee_data"] = {}
            data["payee_data"]["compliance"] = json_dict.pop("compliance")

        if "paymentInfo" in json_dict and "converted" not in json_dict:
            # pylint: disable=protected-access
            data["payment_info"] = PayReqResponsePaymentInfo(
                **PayReqResponsePaymentInfo._from_dict(json_dict.pop("paymentInfo"))
            )

        data["uma_major_version"] = 1 if is_uma_v1 else 0 if is_uma_v0 else None

        # Assert that v1 responses are properly signed.
        if is_uma_v1:
            compliance = PayReqResponseCompliance.from_payee_data(data["payee_data"])
            if not compliance:
                raise ValueError(
                    "UMA v1 responses must include compliance data in payee_data."
                )
            if (
                not compliance.signature
                or not compliance.signature_nonce
                or not compliance.signature_timestamp
            ):
                raise ValueError(
                    "UMA v1 responses must include signature data in compliance."
                )

        return data



from typing import List
from cryptography import x509
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from uma.exceptions import InvalidRequestException
from uma.generated.errors import ErrorCode


def get_pubkey(cert: x509.Certificate) -> bytes:
    # The last 65 bytes of the DER-encoded public key are the uncompressed public key.
    pubkey = cert.public_key()

    if not isinstance(pubkey, ec.EllipticCurvePublicKey):
        raise InvalidRequestException(
            "Public key is not an Elliptic Curve public key.",
            ErrorCode.CERT_CHAIN_INVALID,
        )
    if pubkey.curve.name != "secp256k1":
        raise InvalidRequestException(
            "Public key is not a valid SECP256K1 public key.",
            ErrorCode.CERT_CHAIN_INVALID,
        )

    return pubkey.public_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )[-65:]


def get_x509_certs(cert_chain: str) -> List[x509.Certificate]:
    try:
        return x509.load_pem_x509_certificates(cert_chain.encode())
    except ValueError as exc:
        raise InvalidRequestException(
            "Unable to parse certificate as valid X.509.", ErrorCode.CERT_CHAIN_INVALID
        ) from exc



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from dataclasses import dataclass
from typing import Any, Dict, Optional

from uma.JSONable import JSONable
from uma.protocol.counterparty_data import CounterpartyDataOptions
from uma.protocol.payer_data import PayerData


@dataclass
class PayRequest(JSONable):
    currency_code: Optional[str]
    """
    The currency code for the currency that the receiver will receive for this payment.
    """

    amount: int
    """
    The amount of the payment in the currency specified by `currency_code`. This amount is
    in the smallest unit of the specified currency (e.g. cents for USD).
    """

    payer_data: Optional[PayerData]
    """
    The data about the payer that the sending VASP must provide in order to send a payment.
    This was requested by the receiver in the lnulp response. See LUD-18.
    """

    requested_payee_data: Optional[CounterpartyDataOptions] = None
    """
    The data about the receiver that the sending VASP would like to know from the receiver.
    See LUD-22.
    """

    comment: Optional[str] = None
    """
    A comment that the sender would like to include with the payment. This can only be included
    if the receiver included the `commentAllowed` field in the lnurlp response. The length of
    the comment must be less than or equal to the value of `commentAllowed`.
    """

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {"currency_code": "currency"}

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        data = super()._from_dict(json_dict)
        amount = data.get("amount")
        if isinstance(amount, str):
            data["amount"] = int(amount)
        return data



from dataclasses import dataclass
from typing import Optional

import bech32

from uma.protocol.counterparty_data import (
    CounterpartyDataOption,
    CounterpartyDataOptions,
)
from uma.protocol.kyc_status import KycStatus
from uma.tlv_utils import TLVCodable, ByteCodable


@dataclass
class InvoiceCurrency(TLVCodable):
    # code is the ISO 4217 (if applicable) currency code (eg. "USD"). For cryptocurrencies, this will  be a ticker
    # symbol, such as BTC for Bitcoin.
    code: str

    # name is the full display name of the currency (eg. US Dollars).
    name: str

    # symbol is the symbol used to represent the currency (eg. $).
    symbol: str

    # The number of digits after the decimal point for display on the sender side
    decimals: int

    def __init__(
        self, code: str = "", name: str = "", symbol: str = "", decimals: int = 0
    ) -> None:
        self.code = code
        self.name = name
        self.symbol = symbol
        self.decimals = decimals

    @classmethod
    def tlv_map(cls) -> dict:
        return {
            "code": 0,
            "name": 1,
            "symbol": 2,
            "decimals": 3,
        }


@dataclass
class InvoiceCounterpartyDataOptions(ByteCodable):
    options: CounterpartyDataOptions

    def to_bytes(self) -> bytes:
        pairs = []
        for key, value in self.options.items():
            result = key + ":"
            if value.mandatory:
                result += "1"
            else:
                result += "0"
            pairs.append(result)
        pairs.sort()
        return ",".join(pairs).encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> "InvoiceCounterpartyDataOptions":
        options = {}
        for pair in data.decode().split(","):
            key, value = pair.split(":")
            options[key] = CounterpartyDataOption(mandatory=value == "1")
        return cls(options)


@dataclass
class Invoice(TLVCodable):
    # Receiving UMA address
    receiver_uma: str

    # Invoice UUID Served as both the identifier of the UMA invoice, and the validation of proof of payment.
    invoice_uuid: str

    # The amount of invoice to be paid in the smalest unit of the ReceivingCurrency.
    amount: int

    # The currency of the invoice
    receving_currency: InvoiceCurrency

    # The unix timestamp the UMA invoice expires
    expiration: int

    # Indicates whether the VASP is a financial institution that requires travel rule information.
    is_subject_to_travel_rule: bool

    # RequiredPayerData the data about the payer that the sending VASP must provide in order to send a payment.
    required_payer_data: Optional[InvoiceCounterpartyDataOptions]

    # UmaVersion is a list of UMA versions that the VASP supports for this transaction. It should be
    # containing the highest minor version of each major version it supported, separated by commas.
    uma_versions: str

    # CommentCharsAllowed is the number of characters that the sender can include in the comment field of the pay request.
    comment_chars_allowed: Optional[int]

    # The sender's UMA address. If this field presents, the UMA invoice should directly go to the sending VASP instead of showing in other formats.
    sender_uma: Optional[str]

    # The maximum number of times the invoice can be paid
    max_num_payments: Optional[int]

    # KYC status of the receiver, default is verified.
    kyc_status: Optional[KycStatus]

    # The callback url that the sender should send the PayRequest to.
    callback: str

    # The signature of the UMA invoice
    signature: Optional[bytes]

    def __init__(
        self,
        receiver_uma: str = "",
        invoice_uuid: str = "",
        amount: int = 0,
        receving_currency: InvoiceCurrency = InvoiceCurrency(),
        expiration: int = 0,
        is_subject_to_travel_rule: bool = False,
        required_payer_data: Optional[InvoiceCounterpartyDataOptions] = None,
        uma_versions: str = "",
        comment_chars_allowed: Optional[int] = None,
        sender_uma: Optional[str] = None,
        max_num_payments: Optional[int] = None,
        kyc_status: Optional[KycStatus] = None,
        callback: str = "",
        signature: Optional[bytes] = None,
    ) -> None:
        self.receiver_uma = receiver_uma
        self.invoice_uuid = invoice_uuid
        self.amount = amount
        self.receving_currency = receving_currency
        self.expiration = expiration
        self.is_subject_to_travel_rule = is_subject_to_travel_rule
        self.required_payer_data = required_payer_data
        self.uma_versions = uma_versions
        self.comment_chars_allowed = comment_chars_allowed
        self.sender_uma = sender_uma
        self.max_num_payments = max_num_payments
        self.kyc_status = kyc_status
        self.callback = callback
        self.signature = signature

    @classmethod
    def tlv_map(cls) -> dict:
        return {
            "receiver_uma": 0,
            "invoice_uuid": 1,
            "amount": 2,
            "receving_currency": 3,
            "expiration": 4,
            "is_subject_to_travel_rule": 5,
            "required_payer_data": 6,
            "uma_versions": 7,
            "comment_chars_allowed": 8,
            "sender_uma": 9,
            "max_num_payments": 10,
            "kyc_status": 11,
            "callback": 12,
            "signature": 100,
        }

    def to_bech32_string(self) -> str:
        data = bech32.convertbits(self.to_tlv(), 8, 5)
        if data is None:
            raise ValueError("Failed to convert to bech32")
        return bech32.bech32_encode("uma", data)

    @classmethod
    def from_bech32_string(cls, bech32_str: str) -> "Invoice":
        _, data = bech32.bech32_decode(bech32_str)
        if data is None:
            raise ValueError("Failed to decode bech32")
        tlvs = bech32.convertbits(data, 5, 8)
        if tlvs is None:
            raise ValueError("Failed to convert bits")
        return cls.from_tlv(bytes(tlvs))



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from dataclasses import dataclass
from typing import Any, Dict, Optional

from uma.protocol.counterparty_data import CounterpartyDataOptions
from uma.JSONable import JSONable
from uma.protocol.payer_data import PayerData
from uma.protocol.settlement import SettlementInfo


@dataclass
class PayRequest(JSONable):
    sending_amount_currency_code: Optional[str]
    """
    The currency code of the `amount` field. `None` indicates that `amount` is in the smallest
    unit of the settlement asset. For lightning, this is millisatoshis as in LNURL without LUD-21.
    If this is not `None`, then `amount` is in the smallest unit of
    the specified currency (e.g. cents for USD). This currency code can be any currency which
    the receiver can quote. However, there are two most common scenarios for UMA:

    1. If the sender wants the receiver wants to receive a specific amount in their receiving
    currency, then this field should be the same as `receiving_currency_code`. This is useful
    for cases where the sender wants to ensure that the receiver receives a specific amount
    in that destination currency, regardless of the exchange rate, for example, when paying
    for some goods or services in a foreign currency.

    2. If the sender has a specific amount in their own currency that they would like to send,
    then this field should be left as `None` to indicate that the amount is in the smallest
    unit of the settlement asset (ie. msats by default).
    This will lock the sent amount on the sender side, and the receiver will receive the
    equivalent amount in their receiving currency. NOTE: In this scenario, the sending VASP
    *should not* pass the sending currency code here, as it is not relevant to the receiver.
    Rather, by specifying an invoice amount in the settlement asset (for example, msats for
    lightning), the sending VASP can ensure that their user will be sending a fixed amount,
    regardless of the exchange rate on the receiving side.
    """

    receiving_currency_code: Optional[str]
    """
    The currency code for the currency that the receiver will receive for this payment.
    """

    amount: int
    """
    The amount of the payment in the currency specified by `currency_code`. This amount is
    in the smallest unit of the specified currency (e.g. cents for USD).
    """

    payer_data: Optional[PayerData]
    """
    The data about the payer that the sending VASP must provide in order to send a payment.
    This was requested by the receiver in the lnulp response. See LUD-18.
    """

    requested_payee_data: Optional[CounterpartyDataOptions]
    """
    The data about the receiver that the sending VASP would like to know from the receiver.
    See LUD-22.
    """

    comment: Optional[str] = None
    """
    A comment that the sender would like to include with the payment. This can only be included
    if the receiver included the `commentAllowed` field in the lnurlp response. The length of
    the comment must be less than or equal to the value of `commentAllowed`.
    """

    settlement_info: Optional[SettlementInfo] = None
    """
    Settlement information including the layer and asset chosen by the sender.
    Must be one of the options provided by the receiver in the lnurlp response.
    If not specified, defaults to Lightning with BTC.
    """

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {
            "requested_payee_data": "payeeData",
            "settlement_info": "settlement",
        }

    def to_dict(self) -> Dict[str, Any]:
        result_dict = super().to_dict()
        sending_currency = (
            result_dict.pop("sendingAmountCurrencyCode")
            if "sendingAmountCurrencyCode" in result_dict
            else None
        )
        if self.receiving_currency_code is not None:
            receiving_currency = result_dict.pop("receivingCurrencyCode")
            result_dict["convert"] = receiving_currency
        result_dict["amount"] = (
            f"{result_dict['amount']}.{sending_currency}"
            if sending_currency is not None
            else f"{result_dict['amount']}"
        )
        return result_dict

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        data = super()._from_dict(json_dict)
        if "convert" in json_dict:
            data["receiving_currency_code"] = json_dict.pop("convert")
        if "amount" in data:
            amount = str(data.pop("amount"))
            if "." in amount:
                [amount, currency] = amount.split(".")
                data["amount"] = int(amount)
                data["sending_amount_currency_code"] = currency
            else:
                data["amount"] = int(amount)
        return data



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, List
from urllib.parse import urlencode

from uma.exceptions import InvalidRequestException
from uma.generated.errors import ErrorCode
from uma.protocol.backing_signature import BackingSignature
from uma.signing_utils import sign_payload
from uma.type_utils import none_throws
from uma.urls import is_domain_local


@dataclass
class LnurlpRequest:
    receiver_address: str
    """
    The UMA or Lightning address of the receiver.
    """

    nonce: Optional[str]
    """
    A random string included in the signature payload to prevent replay attacks.
    """

    signature: Optional[str]
    """
    DER-encoded signature from the sending VASP.
    """

    is_subject_to_travel_rule: Optional[bool]
    """
    Whether the sending VASP is subject to travel rule regulations.
    """

    vasp_domain: Optional[str]
    """
    The domain of the sending VASP.
    """

    timestamp: Optional[datetime]
    """
    The time at which the request was made.
    """

    uma_version: Optional[str]
    """
    The version of the UMA protocol that the sender is using.
    """

    backing_signatures: Optional[List[BackingSignature]] = None
    """
    List of backing VASP signatures.
    """

    def encode_to_url(self) -> str:
        try:
            [identifier, host] = self.receiver_address.split("@")
        except ValueError as ex:
            raise InvalidRequestException(
                f"invalid receiver address {self.receiver_address}.",
                ErrorCode.INTERNAL_ERROR,
            ) from ex

        scheme = "http" if is_domain_local(host) else "https"
        base_url = f"{scheme}://{host}/.well-known/lnurlp/{identifier}"
        backing_signatures = [
            f"{sig.domain}:{sig.signature}" for sig in (self.backing_signatures or [])
        ]

        if not self.is_uma_request():
            return base_url
        params = {
            "signature": self.signature,
            "vaspDomain": self.vasp_domain,
            "nonce": self.nonce,
            "isSubjectToTravelRule": str(self.is_subject_to_travel_rule).lower(),
            "timestamp": int(none_throws(self.timestamp).timestamp()),
            "umaVersion": self.uma_version,
        }
        if backing_signatures:
            params["backingSignatures"] = ",".join(backing_signatures)
        return f"{base_url}?{urlencode(params)}"

    def signable_payload(self) -> bytes:
        if not self.nonce or not self.timestamp:
            raise InvalidRequestException(
                "nonce and timestamp are required for signing. This is not an UMA request.",
                ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
            )
        signable = "|".join(
            [self.receiver_address, self.nonce, str(int(self.timestamp.timestamp()))]
        )
        return signable.encode("utf8")

    def is_uma_request(self) -> bool:
        return (
            self.uma_version is not None
            and self.signature is not None
            and self.nonce is not None
            and self.timestamp is not None
            and self.is_subject_to_travel_rule is not None
        )

    def append_backing_signature(self, signing_private_key: bytes, domain: str) -> None:
        """
        Appends a backing signature to the lnurlp request.

        Args:
            signing_private_key: The private key of the backing VASP which is used to sign the payload.
            domain: The domain of the backing VASP that produced the signature. Public keys for this VASP
            will be fetched from this domain at /.well-known/lnurlpubkey and used to verify the signature.
        """
        payload = self.signable_payload()
        backing_signature = sign_payload(payload, signing_private_key)
        if self.backing_signatures is None:
            self.backing_signatures = []
        self.backing_signatures.append(
            BackingSignature(domain=domain, signature=backing_signature)
        )



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

import json
from dataclasses import dataclass
from functools import total_ordering
from typing import List, Optional

BACK_COMPATIBLE_VERSIONS = ["0.3"]
MAJOR_VERSION = 1
MINOR_VERSION = 3
UMA_PROTOCOL_VERSION = f"{MAJOR_VERSION}.{MINOR_VERSION}"


@total_ordering
@dataclass
class ParsedVersion:
    major: int
    minor: int

    @classmethod
    def load(cls, version: str) -> "ParsedVersion":
        [major, minor] = version.split(".")
        return ParsedVersion(major=int(major), minor=int(minor))

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}"

    def __lt__(self, other: "ParsedVersion") -> bool:
        return self.major < other.major or (
            self.major == other.major and self.minor < other.minor
        )


def get_supported_major_versions() -> List[int]:
    """
    In the future, we may want to support multiple major versions in the same SDK,
    but for now, this keeps things simple.
    """
    return [MAJOR_VERSION] + [
        ParsedVersion.load(v).major for v in BACK_COMPATIBLE_VERSIONS
    ]


def get_highest_supported_version_for_major_version(
    major_version: int,
) -> Optional[ParsedVersion]:
    """
    Note that this also only supports a single major version for now. If we support
    more than one major version in the future, we'll need to change this.
    """
    if major_version == MAJOR_VERSION:
        return ParsedVersion.load(UMA_PROTOCOL_VERSION)

    for version in BACK_COMPATIBLE_VERSIONS:
        parsed_version = ParsedVersion.load(version)
        if parsed_version.major == major_version:
            return parsed_version

    return None


def select_highest_supported_version(
    other_vasp_supported_major_versions: List[int],
) -> Optional[str]:
    supported_major_versions = get_supported_major_versions()

    highest_version = None
    for other_vasp_major_version in other_vasp_supported_major_versions:
        if other_vasp_major_version not in supported_major_versions:
            continue

        if not highest_version or other_vasp_major_version > highest_version.major:
            highest_version = get_highest_supported_version_for_major_version(
                other_vasp_major_version
            )

    return str(highest_version) if highest_version else None


def select_lower_version(version_1: str, version_2: str) -> str:
    return (
        version_1
        if ParsedVersion.load(version_1) < ParsedVersion.load(version_2)
        else version_2
    )


def is_version_supported(version: str) -> bool:
    parsed_version = ParsedVersion.load(version)
    return parsed_version.major in get_supported_major_versions()


def get_supported_major_versions_from_error(error_json: str) -> List[int]:
    error = json.loads(error_json)
    return error["supportedMajorVersions"]



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved
import json
import random
import re
from dataclasses import replace
from datetime import datetime, timezone
from math import floor
from typing import Dict, List, Optional
from urllib.parse import parse_qs, unquote, urlparse
from uuid import uuid4

import requests
from aiohttp import ClientSession
from coincurve.ecdsa import cdata_to_der, der_to_cdata, signature_normalize
from coincurve.keys import PublicKey
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from ecies import encrypt

from uma.cert_utils import get_pubkey, get_x509_certs
from uma.exceptions import (
    InvalidCurrencyException,
    InvalidRequestException,
    InvalidSignatureException,
    UnsupportedVersionException,
)
from uma.generated.errors import ErrorCode
from uma.nonce_cache import INonceCache
from uma.protocol.backing_signature import BackingSignature
from uma.protocol.counterparty_data import (
    CounterpartyDataKeys,
    CounterpartyDataOption,
    CounterpartyDataOptions,
)
from uma.protocol.currency import Currency
from uma.protocol.invoice import (
    Invoice,
    InvoiceCounterpartyDataOptions,
    InvoiceCurrency,
)
from uma.protocol.kyc_status import KycStatus
from uma.protocol.lnurlp_request import LnurlpRequest
from uma.protocol.lnurlp_response import LnurlComplianceResponse, LnurlpResponse
from uma.protocol.payer_data import (
    CompliancePayerData,
    PayerData,
    compliance_from_payer_data,
    create_payer_data,
)
from uma.protocol.payreq import PayRequest
from uma.protocol.payreq_response import (
    PayReqResponse,
    PayReqResponseCompliance,
    PayReqResponsePaymentInfo,
)
from uma.protocol.settlement import SettlementInfo, SettlementOption
from uma.protocol.post_tx_callback import (
    PostTransactionCallback,
    UtxoWithAmount,
    TransactionStatus,
)
from uma.protocol.pubkey_response import PubkeyResponse
from uma.public_key_cache import IAsyncPublicKeyCache, IPublicKeyCache
from uma.signing_utils import sign_payload
from uma.type_utils import none_throws
from uma.uma_invoice_creator import IUmaInvoiceCreator
from uma.urls import is_domain_local
from uma.version import (
    UMA_PROTOCOL_VERSION,
    ParsedVersion,
    get_supported_major_versions,
    is_version_supported,
    select_lower_version,
)


def fetch_public_key_for_vasp(
    vasp_domain: str, cache: IPublicKeyCache
) -> PubkeyResponse:
    public_key = cache.fetch_public_key_for_vasp(vasp_domain)
    if public_key:
        return public_key

    scheme = "http://" if is_domain_local(vasp_domain) else "https://"
    url = scheme + vasp_domain + "/.well-known/lnurlpubkey"
    try:
        response = _run_http_get(url)
    except Exception as ex:  # pylint: disable=broad-except
        raise InvalidRequestException(
            f"Unable to fetch pubkey from {vasp_domain}. Make sure the vasp domain is correct.",
            ErrorCode.COUNTERPARTY_PUBKEY_FETCH_ERROR,
        ) from ex
    public_key = PubkeyResponse.from_json(response)
    cache.add_public_key_for_vasp(vasp_domain, public_key)
    return public_key


async def gen_fetch_public_key_for_vasp(
    vasp_domain: str, cache: IAsyncPublicKeyCache
) -> PubkeyResponse:
    public_key = await cache.gen_fetch_public_key_for_vasp(vasp_domain)
    if public_key:
        return public_key

    scheme = "http://" if is_domain_local(vasp_domain) else "https://"
    url = scheme + vasp_domain + "/.well-known/lnurlpubkey"
    try:
        response_text = await _gen_run_http_get(url)
    except Exception as ex:
        raise InvalidRequestException(
            f"Unable to fetch pubkey from {vasp_domain}. Make sure the vasp domain is correct.",
            ErrorCode.COUNTERPARTY_PUBKEY_FETCH_ERROR,
        ) from ex

    public_key = PubkeyResponse.from_json(response_text)
    await cache.gen_add_public_key_for_vasp(vasp_domain, public_key)
    return public_key


def create_pubkey_response(
    signing_cert_chain: str,
    encryption_cert_chain: str,
    expiration_timestamp: Optional[datetime] = None,
) -> PubkeyResponse:
    """
    Creates a pubkey response.

    Args:
        signing_cert_chain: the chain of signing certificates in PEM format.
            The order of the certificates should be from the leaf to the root.
        encryption_cert_chain: the chain of encryption certificates in PEM format.
            The order of the certificates should be from the leaf to the root.
        expiration_timestamp: the expiration timestamp of the public keys.
    """
    signing_certs = get_x509_certs(signing_cert_chain)
    encryption_certs = get_x509_certs(encryption_cert_chain)
    return PubkeyResponse(
        signing_certs,
        encryption_certs,
        get_pubkey(signing_certs[0]),
        get_pubkey(encryption_certs[0]),
        expiration_timestamp,
    )


def parse_pubkey_response(response: str) -> PubkeyResponse:
    return PubkeyResponse.from_json(response)


def _run_http_get(url: str) -> str:
    session = requests.session()
    response = session.get(url=url)
    response.raise_for_status()
    session.close()
    return response.text


async def _gen_run_http_get(url: str) -> str:
    async with ClientSession() as session:
        async with session.get(url) as response:
            response.raise_for_status()
            return await response.text()


def generate_nonce() -> str:
    return str(random.randint(0, 0xFFFFFFFF))


def verify_pay_request_signature(
    request: PayRequest, other_vasp_pubkeys: PubkeyResponse, nonce_cache: INonceCache
) -> None:
    if not request.payer_data:
        raise InvalidRequestException(
            "UMA requires payer data in request. For regular LNURL requests, "
            + "payer data is optional and signatures should not be checked.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )
    compliance_data = compliance_from_payer_data(request.payer_data)
    if not compliance_data:
        raise InvalidRequestException(
            "Missing compliance data in request",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    nonce_cache.check_and_save_nonce(
        compliance_data.signature_nonce,
        _parse_timestamp(compliance_data.signature_timestamp),
    )
    _verify_signature(
        request.signable_payload(),
        compliance_data.signature,
        other_vasp_pubkeys.get_signing_pubkey(),
    )


def verify_pay_request_backing_signatures(
    request: PayRequest, cache: IPublicKeyCache
) -> None:
    if not request.payer_data:
        raise InvalidRequestException(
            "UMA requires payer data in request. For regular LNURL requests, "
            + "payer data is optional and signatures should not be checked.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )
    compliance = compliance_from_payer_data(request.payer_data)
    if not compliance:
        raise InvalidRequestException(
            "Missing compliance data in request",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    backing_signatures = compliance.backing_signatures
    if backing_signatures:
        payload = request.signable_payload()
        for backing_sig in backing_signatures:
            backing_vasp_pubkeys = fetch_public_key_for_vasp(backing_sig.domain, cache)
            _verify_signature(
                payload,
                backing_sig.signature,
                backing_vasp_pubkeys.get_signing_pubkey(),
            )


def _verify_signature(payload: bytes, signature: str, signing_pubkey: bytes) -> None:
    """
    Verifies the signature of the uma request.

    Args:
        payload: the raw payload.
        signature: the hex-encoded signature.
        other_vasp_pubkey: the bytes of the signing public key of the VASP who signed the payload.
    """
    key = _load_public_key(signing_pubkey)
    try:
        _, normalized_signature = signature_normalize(
            der_to_cdata(bytes.fromhex(signature))
        )
        did_verify = key.verify(
            signature=cdata_to_der(normalized_signature),
            message=payload,
        )
        if not did_verify:
            raise InvalidSignature()
    except (ValueError, InvalidSignature) as ex:
        raise InvalidSignatureException() from ex


def _load_public_key(key: bytes) -> PublicKey:
    try:
        return PublicKey(key)
    except ValueError:
        der_key = serialization.load_der_public_key(key)
        return PublicKey(
            der_key.public_bytes(
                encoding=serialization.Encoding.X962,
                format=serialization.PublicFormat.UncompressedPoint,
            )
        )


def _encrypt_travel_rule_info(
    travel_rule_info: str, receiver_encryption_pubkey: bytes
) -> str:
    public_key = _load_public_key(receiver_encryption_pubkey)
    return encrypt(public_key.format(), travel_rule_info.encode()).hex()


def create_compliance_payer_data(
    receiver_encryption_pubkey: bytes,
    signing_private_key: bytes,
    payer_identifier: str,
    travel_rule_info: Optional[str],
    payer_kyc_status: KycStatus,
    payer_utxos: List[str],
    payer_node_pubkey: Optional[str],
    utxo_callback: str,
    travel_rule_format: Optional[str] = None,
) -> CompliancePayerData:
    timestamp = int(datetime.now(timezone.utc).timestamp())
    nonce = generate_nonce()
    encrypted_travel_rule_info = (
        _encrypt_travel_rule_info(travel_rule_info, receiver_encryption_pubkey)
        if travel_rule_info
        else None
    )
    payload = "|".join([payer_identifier, nonce, str(timestamp)])
    signature = sign_payload(payload.encode(), signing_private_key)
    return CompliancePayerData(
        kyc_status=payer_kyc_status,
        utxos=payer_utxos,
        node_pubkey=payer_node_pubkey,
        encrypted_travel_rule_info=encrypted_travel_rule_info,
        travel_rule_format=travel_rule_format,
        signature=signature,
        signature_nonce=nonce,
        signature_timestamp=timestamp,
        utxo_callback=utxo_callback,
    )


def create_pay_request(
    receiving_currency_code: str,
    amount: int,
    is_amount_in_receiving_currency: bool,
    payer_identifier: str,
    uma_major_version: int,
    payer_name: Optional[str],
    payer_email: Optional[str],
    payer_compliance: Optional[CompliancePayerData],
    requested_payee_data: Optional[CounterpartyDataOptions] = None,
    comment: Optional[str] = None,
    invoice_uuid: Optional[str] = None,
    settlement_info: Optional[SettlementInfo] = None,
) -> PayRequest:
    """
    Creates a payreq request object.

    Args:
        receiving_currency_code: The code of the currency that the receiver will receive for this
            payment.
        amount: The amount that the receiver will receive in either the smallest unit of the
            receiving currency (if is_amount_in_receiving_currency is True), or in the smallest unit
            of the settlement asset (if false).
        is_amount_in_receiving_currency: Whether the amount field is specified in the smallest unit
            of the receiving currency or in the smallest unit of the settlement asset (if false).
        payer_identifier: The UMA address of the sender. For example, $alice@vasp.com.
        uma_major_version: The major version of the UMA protocol that this currency adheres to.
            If non-UMA, this version is still relevant for which LUD-21 spec to follow. For the older
            LUD-21 spec, this should be 0. For the newer LUD-21 spec, this should be 1.
        payer_name: The name of the sender if requested by the receiver.
        payer_email: The email of the sender if requested by the receiver.
        payer_compliance: The compliance data of the sender. This is REQUIRED for UMA payments,
            but not for regular LNURL payments.
        requested_payee_data: the additional data about the payee which is requested by the sending
            VASP, if any.
        comment: A comment that the sender wants to add to the payment. This can only be included
            if the receiver included the `commentAllowed` field in the lnurlp response. The length of
            the comment must be less than or equal to the value of `commentAllowed`.
        invoice_uuid: if the PayRequest is for an UMA invoice, this should be the UUID of the invoice.
        settlement_info: The settlement info chosen by the sender for this payment.
    """
    payer_data = create_payer_data(
        identifier=payer_identifier,
        name=payer_name,
        email=payer_email,
        compliance=payer_compliance,
    )
    return create_pay_request_with_payer_data(
        receiving_currency_code=receiving_currency_code,
        amount=amount,
        is_amount_in_receiving_currency=is_amount_in_receiving_currency,
        payer_identifier=payer_identifier,
        uma_major_version=uma_major_version,
        payer_compliance=payer_compliance,
        payer_data=payer_data,
        requested_payee_data=requested_payee_data,
        comment=comment,
        invoice_uuid=invoice_uuid,
        settlement_info=settlement_info,
    )


def create_pay_request_with_payer_data(
    receiving_currency_code: str,
    amount: int,
    is_amount_in_receiving_currency: bool,
    payer_identifier: str,
    uma_major_version: int,
    payer_compliance: Optional[CompliancePayerData],
    payer_data: Optional[PayerData] = None,
    requested_payee_data: Optional[CounterpartyDataOptions] = None,
    comment: Optional[str] = None,
    invoice_uuid: Optional[str] = None,
    settlement_info: Optional[SettlementInfo] = None,
) -> PayRequest:
    """
    Creates a payreq request object using the provided payer data.

    Args:
        receiving_currency_code: The code of the currency that the receiver will receive for this
            payment.
        amount: The amount that the receiver will receive in either the smallest unit of the
            receiving currency (if is_amount_in_receiving_currency is True), or in the smallest
            unit of the settlement asset (if false).
        is_amount_in_receiving_currency: Whether the amount field is specified in the smallest unit
            of the receiving currency or in the smallest unit of the settlement asset (if false).
        payer_identifier: The UMA address of the sender. For example, $alice@vasp.com.
        uma_major_version: The major version of the UMA protocol that this currency adheres to.
            If non-UMA, this version is still relevant for which LUD-21 spec to follow. For the older
            LUD-21 spec, this should be 0. For the newer LUD-21 spec, this should be 1.
        payer_compliance: The compliance data of the sender. This is REQUIRED for UMA payments,
            but not for regular LNURL payments.
        payer_data: The payer data of the sender. This is REQUIRED for UMA payments. You should
            ensure that the payer data provided meets the minimum requirements of the counterparty
            (as specified in the LNURLp response under the mandatory requested_payee_data).
        requested_payee_data: the additional data about the payee which is requested by the sending
            VASP, if any.
        comment: A comment that the sender wants to add to the payment. This can only be included
            if the receiver included the `commentAllowed` field in the lnurlp response. The length of
            the comment must be less than or equal to the value of `commentAllowed`.
        invoice_uuid: if the PayRequest is for an UMA invoice, this should be the UUID of the invoice.
        settlement_info: The settlement info chosen by the sender for this payment.
    """
    if uma_major_version == 0 and not is_amount_in_receiving_currency:
        raise InvalidRequestException(
            "UMA v0 does not support sender-locked amounts. Please set is_amount_in_receiving_currency to True.",
            ErrorCode.INTERNAL_ERROR,
        )

    sending_currency_code = (
        receiving_currency_code if is_amount_in_receiving_currency else None
    )
    payer_data = payer_data or {}
    if payer_compliance is not None:
        payer_data[CounterpartyDataKeys.COMPLIANCE.value] = payer_compliance.to_dict()
    payer_data[CounterpartyDataKeys.IDENTIFIER.value] = payer_identifier
    return PayRequest(
        receiving_currency_code=receiving_currency_code,
        sending_amount_currency_code=sending_currency_code,
        amount=amount,
        payer_data=payer_data,
        requested_payee_data=requested_payee_data,
        comment=comment,
        uma_major_version=uma_major_version,
        invoice_uuid=invoice_uuid,
        settlement_info=settlement_info,
    )


def parse_pay_request(payload: str) -> PayRequest:
    return PayRequest.from_json(payload)


def create_uma_lnurlp_request_url(
    signing_private_key: bytes,
    receiver_address: str,
    sender_vasp_domain: str,
    is_subject_to_travel_rule: bool,
    uma_version_override: Optional[str] = None,
) -> str:
    """
    Creates a signed uma request URL.

    Args:
        signing_private_key: the private key of the VASP that is sending the payment. This will be used to sign the request.
        receiver_address: the address of the receiver of the payment (i.e. $bob@vasp2).
        sender_vasp_domain: the domain of the VASP that is sending the payment. It will be used by the receiver to fetch the public keys of the sender.
        is_subject_to_travel_rule: whether the sending VASP is a financial institution that requires travel rule information.
    """

    nonce = generate_nonce()
    request = LnurlpRequest(
        receiver_address=receiver_address,
        nonce=nonce,
        signature="",
        is_subject_to_travel_rule=is_subject_to_travel_rule,
        vasp_domain=sender_vasp_domain,
        timestamp=datetime.now(timezone.utc),
        uma_version=uma_version_override or UMA_PROTOCOL_VERSION,
    )
    request.signature = sign_payload(request.signable_payload(), signing_private_key)
    return request.encode_to_url()


def parse_lnurlp_request(url: str) -> LnurlpRequest:
    parsed_url = urlparse(url)
    query = parse_qs(parsed_url.query, keep_blank_values=True)
    signature = query.get("signature", [""])[0] if query.get("signature") else None
    vasp_domain = query.get("vaspDomain", [""])[0] if query.get("vaspDomain") else None
    nonce = query.get("nonce", [""])[0] if query.get("nonce") else None
    timestamp = query.get("timestamp", [""])[0] if query.get("timestamp") else None
    uma_version = query.get("umaVersion", [""])[0] if query.get("umaVersion") else None
    backing_signatures = (
        query.get("backingSignatures", [""])[0]
        if query.get("backingSignatures")
        else None
    )

    if uma_version and not is_version_supported(uma_version):
        raise UnsupportedVersionException(
            unsupported_version=uma_version,
            supported_major_versions=get_supported_major_versions(),
        )

    required_uma_fields = {
        "signature": signature,
        "vasp_domain": vasp_domain,
        "nonce": nonce,
        "timestamp": timestamp,
        "uma_version": uma_version,
    }
    # UMA fields are all or nothing. If any are present, all must be present.
    has_an_uma_field = any(required_uma_fields.values())
    has_all_uma_fields = all(required_uma_fields.values())
    if has_an_uma_field and not has_all_uma_fields:
        raise InvalidRequestException(
            "Missing uma query parameters: vaspDomain, signature, nonce, uma_version, and timestamp are required.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    paths = parsed_url.path.split("/")
    if len(paths) != 4 or paths[1] != ".well-known" or paths[2] != "lnurlp":
        raise InvalidRequestException(
            "Invalid request path.", ErrorCode.PARSE_LNURLP_REQUEST_ERROR
        )

    if not re.match(r"^[\$a-zA-Z0-9@._\-\+]+$", paths[3]):
        raise InvalidRequestException(
            "Invalid characters in receiver address.",
            ErrorCode.PARSE_LNURLP_REQUEST_ERROR,
        )

    receiver_address = paths[3] + "@" + parsed_url.netloc
    is_subject_to_travel_rule = (
        query.get("isSubjectToTravelRule", [""])[0].lower() == "true"
    )

    backing_signatures_parsed = None
    if backing_signatures:
        pairs = backing_signatures.split(",")
        signatures = []
        for pair in pairs:
            try:
                decoded_pair = unquote(pair)
            except TypeError as ex:
                raise InvalidRequestException(
                    "Invalid backing signature format",
                    ErrorCode.PARSE_LNURLP_REQUEST_ERROR,
                ) from ex

            last_colon_index = decoded_pair.rfind(":")
            if last_colon_index == -1:
                raise InvalidRequestException(
                    "Invalid backing signature format",
                    ErrorCode.PARSE_LNURLP_REQUEST_ERROR,
                )

            signatures.append(
                BackingSignature(
                    domain=decoded_pair[:last_colon_index],
                    signature=decoded_pair[last_colon_index + 1 :],
                )
            )
        backing_signatures_parsed = signatures

    return LnurlpRequest(
        receiver_address=receiver_address,
        nonce=nonce,
        signature=signature,
        is_subject_to_travel_rule=is_subject_to_travel_rule,
        vasp_domain=vasp_domain,
        timestamp=(_parse_timestamp(int(timestamp)) if timestamp else None),
        uma_version=uma_version,
        backing_signatures=backing_signatures_parsed,
    )


def is_uma_lnurlp_query(url: str) -> bool:
    try:
        request = parse_lnurlp_request(url)
        return request.is_uma_request()
    except UnsupportedVersionException:
        # Unsupported versions are still UMA requests. The version negotiation
        # should be handled by the VASP when parsing the request.
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def verify_uma_lnurlp_query_signature(
    request: LnurlpRequest, other_vasp_pubkeys: PubkeyResponse, nonce_cache: INonceCache
) -> None:
    """
    Verifies the primary signature on an uma Lnurlp query based on the public key of the VASP
    making the request.

    Args:
        request: the signed request to verify.
        other_vasp_signing_pubkey: the public key of the VASP making this request in bytes.
    """
    if not request.signature or not request.nonce or not request.timestamp:
        raise InvalidRequestException(
            "Missing uma query parameters: signature, nonce and timestamp are required.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    nonce_cache.check_and_save_nonce(request.nonce, request.timestamp)
    _verify_signature(
        request.signable_payload(),
        none_throws(request.signature),
        other_vasp_pubkeys.get_signing_pubkey(),
    )


def verify_uma_lnurlp_query_backing_signatures(
    request: LnurlpRequest, cache: IPublicKeyCache
) -> None:
    """
    Verifies the backing signatures on an uma Lnurlp query. You may optionally
    call this function after verify_uma_lnurlp_query_signature to verify signatures
    from backing VASPs.

    Args:
        request: the signed request to verify.
        cache: the public key cache to use. You can use the InMemoryPublicKeyCache class,
        or implement your own persistent cache with any storage type.
    """
    if not request.backing_signatures:
        return

    backing_signatures = request.backing_signatures
    if backing_signatures:
        payload = request.signable_payload()
        for backing_sig in backing_signatures:
            backing_vasp_pubkeys = fetch_public_key_for_vasp(backing_sig.domain, cache)
            _verify_signature(
                payload,
                backing_sig.signature,
                backing_vasp_pubkeys.get_signing_pubkey(),
            )


def _add_invoice_uuid_to_metadata(metadata: str, invoice_uuid: str) -> str:
    try:
        data = json.loads(metadata)
        data.append(["text/uma-invoice", invoice_uuid])
        return json.dumps(data)
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON: {e}") from e


def create_pay_req_response(
    request: PayRequest,
    invoice_creator: IUmaInvoiceCreator,
    metadata: str,
    receiving_currency_code: Optional[str],
    receiving_currency_decimals: Optional[int],
    msats_per_currency_unit: Optional[float],
    receiver_fees_msats: Optional[int],
    receiver_node_pubkey: Optional[str],
    utxo_callback: Optional[str],
    payee_identifier: Optional[str],
    signing_private_key: Optional[bytes],
    receiver_utxos: Optional[List[str]] = None,
    payee_data: Optional[PayerData] = None,
    disposable: bool = True,
    success_action: Optional[Dict[str, str]] = None,
) -> PayReqResponse:
    """
    Creates an uma pay request response with an encoded invoice.

    Args:
        request: the uma pay request. This will be used to sign the request.
        invoice_creator: the object that will create the invoice. In practice, this is usually a `services.LightsparkClient`.
        metadata: the metadata that will be added to the invoice's metadata hash field. This should be a JSON array as specified in LUD-06.
        receiving_currency_code: the code of the currency that the receiver will receive for this payment. Required for UMA transactions.
        receiving_currency_decimals: the number of decimal places in the specified currency. For example, USD has 2 decimal
            places. This should align with the decimals field returned for the chosen currency in the LNURLP response. Required for UMA transactions.
        msats_per_currency_unit: milli-satoshis per the smallest unit of the specified currency. This rate is committed to by the receiving VASP until the
            invoice expires. Required for UMA transactions.
        receiver_fees_msats: the fees charged (in millisats) by the receiving VASP to convert to the target currency. This is separate from the
            conversion rate. Required for UMA transactions.
        receiver_node_pubkey: the public key of the receiver node.
        utxo_callback: the URL that the receiving VASP will call to send UTXOs of the channel that the receiver used to receive the payment once it completes.
        payee_identifier: the identifier of the receiver. For example, $bob@vasp2.com. Required for UMA transactions.
        signing_private_key: the private key of the VASP that is receiving the payment. This will be used to sign the request. Required for UMA transactions.
        receiver_utxos: the list of UTXOs of the receiver's channels that might be used to fund the payment.
        payee_data: the additional data about the payee which was requested in the pay request by the sending VASP, if any.
        disposable: This field may be used by a WALLET to decide whether the initial LNURL link will be stored locally for later reuse or erased.
            If disposable is null, it should be interpreted as true, so if SERVICE intends its LNURL links to be stored it must return
            `disposable: false`. UMA should never return `disposable: false`. See LUD-11.
        success_action: an action that the wallet should take once the payment is complete. See LUD-09
    """
    if (
        request.sending_amount_currency_code
        and request.sending_amount_currency_code != receiving_currency_code
    ):
        raise InvalidCurrencyException(
            "The sending currency code in the pay request does not match the receiving currency code."
        )
    required_uma_fields = {
        "receiving_currency_code": receiving_currency_code,
        "receiving_currency_decimals": receiving_currency_decimals,
        "msats_per_currency_unit": msats_per_currency_unit,
        "receiver_fees_msats": receiver_fees_msats,
        "payee_identifier": payee_identifier,
        "signing_private_key": signing_private_key,
    }
    if request.is_uma_request():
        for field, value in required_uma_fields.items():
            if value is None:
                raise InvalidRequestException(
                    f"Missing required field {field} for UMA request.",
                    ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
                )

    sending_currency = request.sending_amount_currency_code
    amount_msats = (
        request.amount
        if sending_currency is None
        or msats_per_currency_unit is None
        or receiver_fees_msats is None
        else request.amount * msats_per_currency_unit + receiver_fees_msats
    )
    receiving_amount = (
        request.amount
        if sending_currency is not None
        or msats_per_currency_unit is None
        or receiver_fees_msats is None
        else floor((request.amount - receiver_fees_msats) / msats_per_currency_unit)
    )

    if request.invoice_uuid:
        metadata = _add_invoice_uuid_to_metadata(metadata, request.invoice_uuid)

    # Add payer data to the metadata, per LUD-18.
    if request.payer_data:
        metadata += json.dumps(request.payer_data)

    encoded_invoice = invoice_creator.create_invoice_for_settlement_layer(
        amount_units=round(amount_msats),
        metadata=metadata,
        receiver_identifier=payee_identifier,
        settlement_info=request.settlement_info,
    )
    payer_identifier = (
        request.payer_data[CounterpartyDataKeys.IDENTIFIER.value]
        if request.payer_data
        else None
    )
    if not payer_identifier and request.is_uma_request():
        raise InvalidRequestException(
            "Missing payer identifier in request",
            ErrorCode.MISSING_MANDATORY_PAYER_DATA,
        )
    if request.is_uma_request():
        payee_data = payee_data or {}
        payee_data[CounterpartyDataKeys.COMPLIANCE.value] = (
            _create_compliance_payee_data(
                signing_private_key=none_throws(signing_private_key),
                payer_identifier=none_throws(payer_identifier),
                payee_identifier=none_throws(payee_identifier),
                receiver_utxos=receiver_utxos or [],
                receiver_node_pubkey=receiver_node_pubkey,
                utxo_callback=utxo_callback or "",
            ).to_dict()
        )
        payee_data[CounterpartyDataKeys.IDENTIFIER.value] = payee_identifier
    return PayReqResponse(
        encoded_invoice=encoded_invoice,
        routes=[],
        payee_data=payee_data,
        payment_info=(
            PayReqResponsePaymentInfo(
                amount=receiving_amount,
                currency_code=none_throws(receiving_currency_code),
                decimals=none_throws(receiving_currency_decimals),
                multiplier=none_throws(msats_per_currency_unit),
                exchange_fees_msats=none_throws(receiver_fees_msats),
            )
            if receiving_currency_code is not None
            and msats_per_currency_unit is not None
            else None
        ),
        disposable=disposable,
        success_action=success_action,
        uma_major_version=request.uma_major_version,
    )


def create_pay_req_response_for_settlement_layer(
    request: PayRequest,
    invoice_creator: IUmaInvoiceCreator,
    metadata: str,
    receiving_currency_code: Optional[str],
    receiving_currency_decimals: Optional[int],
    multiplier: Optional[float],
    receiver_fee: Optional[int],
    callback_url: Optional[str],
    payee_identifier: Optional[str],
    signing_private_key: Optional[bytes],
    receiver_node_pubkey: Optional[str] = None,
    payee_data: Optional[PayerData] = None,
    disposable: bool = True,
    success_action: Optional[Dict[str, str]] = None,
) -> PayReqResponse:
    """
    Creates a pay request response. This is the recommended function for working
    with alternate settlement layers. For Lightning-only implementations, you can
    still use `create_pay_req_response`.

    Args:
        request: The uma pay request.
        invoice_creator: Creates the payment request for the settlement layer specified
            in the pay request.
        metadata: Metadata to include in the payment request (JSON array per LUD-06).
        receiving_currency_code: Currency code the receiver will receive (e.g., "USD").
        receiving_currency_decimals: Decimal places in the currency (e.g., 2 for USD).
        multiplier: The conversion rate - how many of the smallest units of the settlement
            asset equal one unit of the receiving currency.
            For example:
            - Lightning/BTC: If 1 USD cent = 34,150 millisatoshis, then multiplier = 34150
            - Spark USDC: If 1 USD cent = 100 token units, then multiplier = 100
        receiver_fee: Fees charged by receiver in smallest units of settlement asset
            (e.g. millisatoshis for Lightning/BTC, token units for Spark USDC).
        callback_url: URL to call when payment completes with transaction details.
        payee_identifier: Receiver's UMA address (e.g., $bob@vasp.com). Required for UMA.
        signing_private_key: Receiver VASP's private key for signing. Required for UMA.
        receiver_node_pubkey: the public key of the receiver node. Only relevant when the
            settlement layer is Lightning.
        payee_data: Additional payee data requested by sender in the pay request.
        disposable: Whether the payment link should be single-use (see LUD-11).
        success_action: Action for wallet to take after payment completes (see LUD-09).

    Returns:
        PayReqResponse with the encoded payment request.
    """
    return create_pay_req_response(
        request=request,
        invoice_creator=invoice_creator,
        metadata=metadata,
        receiving_currency_code=receiving_currency_code,
        receiving_currency_decimals=receiving_currency_decimals,
        msats_per_currency_unit=multiplier,
        receiver_fees_msats=receiver_fee,
        utxo_callback=callback_url,
        payee_identifier=payee_identifier,
        signing_private_key=signing_private_key,
        receiver_node_pubkey=receiver_node_pubkey,
        payee_data=payee_data,
        disposable=disposable,
        success_action=success_action,
    )


def _create_compliance_payee_data(
    signing_private_key: bytes,
    payer_identifier: str,
    payee_identifier: str,
    receiver_utxos: List[str],
    receiver_node_pubkey: Optional[str],
    utxo_callback: str,
) -> PayReqResponseCompliance:
    timestamp = int(datetime.now(timezone.utc).timestamp())
    nonce = generate_nonce()
    compliance_payee_data = PayReqResponseCompliance(
        utxos=receiver_utxos,
        utxo_callback=utxo_callback,
        node_pubkey=receiver_node_pubkey,
        signature="",
        signature_nonce=nonce,
        signature_timestamp=timestamp,
    )
    payload = compliance_payee_data.signable_payload(payer_identifier, payee_identifier)
    signature = sign_payload(payload, signing_private_key)
    compliance_payee_data.signature = signature
    return compliance_payee_data


def parse_pay_req_response(payload: str) -> PayReqResponse:
    return PayReqResponse.from_json(payload)


def verify_pay_req_response_signature(
    sender_address: str,
    receiver_address: str,
    response: PayReqResponse,
    other_vasp_pubkeys: PubkeyResponse,
    nonce_cache: INonceCache,
) -> None:
    payee_data = response.payee_data
    if not payee_data:
        raise InvalidRequestException(
            "Missing payee data in response. Cannot verify signature.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )
    compliance_data = response.get_compliance()
    if not compliance_data:
        raise InvalidRequestException(
            "Missing compliance data in response",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    if response.uma_major_version != 1:
        raise InvalidRequestException(
            "Signatures were added to payreq responses in UMA v1. This response is from an UMA v0 receiving VASP.",
            ErrorCode.INTERNAL_ERROR,
        )

    payee_data_identifier = payee_data.get(CounterpartyDataKeys.IDENTIFIER.value)
    if (
        payee_data_identifier is not None
        and payee_data_identifier.lower() != receiver_address.lower()
    ):
        raise InvalidRequestException(
            f"Payee data identifier {payee_data_identifier} does not match receiver address {receiver_address}.",
            ErrorCode.INVALID_SIGNATURE,
        )
    if payee_data_identifier is None:
        payee_data_identifier = receiver_address

    nonce_cache.check_and_save_nonce(
        none_throws(compliance_data.signature_nonce),
        _parse_timestamp(none_throws(compliance_data.signature_timestamp)),
    )
    _verify_signature(
        compliance_data.signable_payload(sender_address, payee_data_identifier),
        none_throws(compliance_data.signature),
        other_vasp_pubkeys.get_signing_pubkey(),
    )


def verify_pay_req_response_backing_signatures(
    response: PayReqResponse,
    cache: IPublicKeyCache,
    sender_address: str,
    receiver_address: str,
) -> None:
    payee_data = response.payee_data
    if not payee_data:
        raise InvalidRequestException(
            "Missing payee data in response. Cannot verify signature.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )
    compliance_data = response.get_compliance()
    if not compliance_data:
        raise InvalidRequestException(
            "Missing compliance data in response",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    if response.uma_major_version != 1:
        raise InvalidRequestException(
            "Signatures were added to payreq responses in UMA v1. This response is from an UMA v0 receiving VASP.",
            ErrorCode.INTERNAL_ERROR,
        )

    payee_data_identifier = payee_data.get(CounterpartyDataKeys.IDENTIFIER.value)
    if (
        payee_data_identifier is not None
        and payee_data_identifier.lower() != receiver_address.lower()
    ):
        raise InvalidRequestException(
            f"Payee data identifier {payee_data_identifier} does not match receiver address {receiver_address}.",
            ErrorCode.INVALID_SIGNATURE,
        )

    backing_signatures = compliance_data.backing_signatures
    if backing_signatures:
        if payee_data_identifier is None:
            payee_data_identifier = receiver_address
        payload = compliance_data.signable_payload(
            sender_address, payee_data_identifier
        )
        for backing_sig in backing_signatures:
            backing_vasp_pubkeys = fetch_public_key_for_vasp(backing_sig.domain, cache)
            _verify_signature(
                payload,
                backing_sig.signature,
                backing_vasp_pubkeys.get_signing_pubkey(),
            )


def create_uma_lnurlp_response(
    request: LnurlpRequest,
    signing_private_key: bytes,
    requires_travel_rule_info: bool,
    callback: str,
    encoded_metadata: str,
    min_sendable_sats: int,
    max_sendable_sats: int,
    payer_data_options: CounterpartyDataOptions,
    currency_options: List[Currency],
    receiver_kyc_status: KycStatus,
    comment_chars_allowed: Optional[int] = None,
    nostr_pubkey: Optional[str] = None,
    settlement_options: Optional[List[SettlementOption]] = None,
) -> LnurlpResponse:
    if not request.is_uma_request():
        raise InvalidRequestException(
            "The request is not a UMA request. Cannot create an UMA response. "
            + "Just create an LnurlpResponse directly instead.",
            ErrorCode.INTERNAL_ERROR,
        )
    uma_version = select_lower_version(
        none_throws(request.uma_version), UMA_PROTOCOL_VERSION
    )
    compliance = _create_signed_lnurlp_compliance_response(
        request=request,
        signing_private_key=signing_private_key,
        is_subject_to_travel_rule=requires_travel_rule_info,
        receiver_kyc_status=receiver_kyc_status,
    )
    _validate_currency_options(currency_options)
    # Ensure correct serializing for UMA V0:
    if ParsedVersion.load(uma_version).major == 0:
        for currency in currency_options:
            currency.uma_major_version = 0

    # compliance and identifier are mandatory fields for UMA
    if CounterpartyDataKeys.COMPLIANCE.value not in payer_data_options:
        payer_data_options[CounterpartyDataKeys.COMPLIANCE.value] = (
            CounterpartyDataOption(mandatory=True)
        )
    if CounterpartyDataKeys.IDENTIFIER.value not in payer_data_options:
        payer_data_options[CounterpartyDataKeys.IDENTIFIER.value] = (
            CounterpartyDataOption(mandatory=True)
        )

    return LnurlpResponse(
        tag="payRequest",
        callback=callback,
        min_sendable=min_sendable_sats * 1000,
        max_sendable=max_sendable_sats * 1000,
        encoded_metadata=encoded_metadata,
        currencies=currency_options,
        required_payer_data=payer_data_options,
        compliance=compliance,
        uma_version=uma_version,
        comment_chars_allowed=comment_chars_allowed,
        nostr_pubkey=nostr_pubkey,
        allows_nostr=nostr_pubkey is not None,
        settlement_options=settlement_options,
    )


def _create_signed_lnurlp_compliance_response(
    request: LnurlpRequest,
    signing_private_key: bytes,
    is_subject_to_travel_rule: bool,
    receiver_kyc_status: KycStatus,
) -> LnurlComplianceResponse:
    timestamp = int(datetime.now(timezone.utc).timestamp())
    nonce = generate_nonce()
    payload = "|".join([request.receiver_address, nonce, str(timestamp)])
    signature = sign_payload(payload.encode(), signing_private_key)
    return LnurlComplianceResponse(
        kyc_status=receiver_kyc_status,
        signature=signature,
        signature_nonce=nonce,
        signature_timestamp=timestamp,
        is_subject_to_travel_rule=is_subject_to_travel_rule,
        receiver_identifier=request.receiver_address,
    )


def parse_lnurlp_response(payload: str) -> LnurlpResponse:
    return LnurlpResponse.from_json(payload)


def verify_uma_lnurlp_response_signature(
    response: LnurlpResponse,
    other_vasp_pubkeys: PubkeyResponse,
    nonce_cache: INonceCache,
) -> None:
    if not response.compliance:
        raise InvalidRequestException(
            "Missing compliance data in response",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    nonce_cache.check_and_save_nonce(
        response.compliance.signature_nonce,
        _parse_timestamp(response.compliance.signature_timestamp),
    )
    _verify_signature(
        response.signable_payload(),
        none_throws(response.compliance).signature,
        other_vasp_pubkeys.get_signing_pubkey(),
    )


def verify_uma_lnurlp_response_backing_signatures(
    response: LnurlpResponse, cache: IPublicKeyCache
) -> None:
    """
    Verifies the backing signatures on an uma Lnurlp query. You may optionally
    call this function after verify_uma_lnurlp_response_signature to verify signatures
    from backing VASPs.

    Args:
        response: the signed response to verify.
        cache: the public key cache to use. You can use the InMemoryPublicKeyCache class,
        or implement your own persistent cache with any storage type.
    """
    if not response.compliance:
        raise InvalidRequestException(
            "Missing compliance data in response",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )

    backing_signatures = response.compliance.backing_signatures
    if not backing_signatures:
        return
    for backing_sig in backing_signatures:
        payload = response.signable_payload()
        backing_vasp_pubkeys = fetch_public_key_for_vasp(backing_sig.domain, cache)
        _verify_signature(
            payload, backing_sig.signature, backing_vasp_pubkeys.get_signing_pubkey()
        )


def _validate_currency_options(currency_options: List[Currency]) -> None:
    for currency in currency_options:
        if currency.millisatoshi_per_unit <= 0:
            raise InvalidCurrencyException(
                f"Invalid currency option {currency.code}. The multiplier must be greater than 0."
            )
        if currency.min_sendable < 0 or currency.max_sendable < 0:
            raise InvalidCurrencyException(
                f"Invalid currency option {currency.code}. The min and max sendable amounts must be greater than 0."
            )
        if currency.min_sendable > currency.max_sendable:
            raise InvalidCurrencyException(
                f"Invalid currency option {currency.code}. The min sendable amount must be less than or equal to the max sendable amount."
            )
        if currency.decimals < 0 or currency.decimals > 8:
            raise InvalidCurrencyException(
                f"Invalid currency option {currency.code}. The number of decimals must be between 0 and 8."
            )


def get_vasp_domain_from_uma_address(uma_address: str) -> str:
    # Gets the domain of the VASP from an uma address.

    [_, domain] = uma_address.split("@")
    return domain


def create_post_transaction_callback(
    utxos: List[UtxoWithAmount],
    vasp_domain: str,
    signing_private_key: bytes,
    transaction_status: Optional[TransactionStatus] = None,
    error_code: Optional[str] = None,
    error_reason: Optional[str] = None,
) -> PostTransactionCallback:
    """
    Creates a signed post transaction callback.

    Args:
        utxos: UTXOs of the channels of the VASP initiating the callback.
        vasp_domain: the domain of the VASP initiating the callback.
        signing_private_key: the private key of the VASP initiating the callback. This will be used to sign the request.
        transaction_status: the status of the transaction.
        error_code: the error code of the transaction. This should be one of the values in [ErrorCode]. Should only be set if transaction_status is FAILED.
        error_reason: the reason for the transaction failure. This should only be set if transaction_status is FAILED.
    """
    timestamp = int(datetime.now(timezone.utc).timestamp())
    nonce = generate_nonce()
    post_transaction_callback = PostTransactionCallback(
        utxos=utxos,
        vasp_domain=vasp_domain,
        signature="",
        signature_nonce=nonce,
        signature_timestamp=timestamp,
        transaction_status=transaction_status,
        error_code=error_code,
        error_reason=error_reason,
    )
    payload = post_transaction_callback.signable_payload()
    signature = sign_payload(payload, signing_private_key)
    post_transaction_callback.signature = signature
    return post_transaction_callback


def parse_post_transaction_callback(payload: str) -> PostTransactionCallback:
    return PostTransactionCallback.from_json(payload)


def verify_post_transaction_callback_signature(
    callback: PostTransactionCallback,
    other_vasp_pubkeys: PubkeyResponse,
    nonce_cache: INonceCache,
) -> None:
    """
    Verifies the signature on a post transaction callback based on the public key of the counterparty

    Args:
        callback: the signed callback to verify.
        other_vasp_signing_pubkey: the public key of the counterparty VASP.
        nonce_cache: the nonce cache used to prevent replay attacks.
    """
    if (
        not callback.signature
        or not callback.signature_nonce
        or not callback.signature_timestamp
    ):
        raise InvalidRequestException(
            "Missing post transaction callback signature, nonce, or timestamp. Is this an UMA v0 callback? If so, don't verify the signature.",
            ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
        )
    nonce_cache.check_and_save_nonce(
        callback.signature_nonce,
        _parse_timestamp(callback.signature_timestamp),
    )
    _verify_signature(
        callback.signable_payload(),
        none_throws(callback.signature),
        other_vasp_pubkeys.get_signing_pubkey(),
    )


def _parse_timestamp(
    timestamp: float,
) -> datetime:
    try:
        return datetime.fromtimestamp(timestamp, timezone.utc)
    except ValueError as ex:
        raise InvalidRequestException(
            "Invalid timestamp in request.", ErrorCode.INVALID_TIMESTAMP
        ) from ex


def create_uma_invoice(
    receiver_uma: str,
    receiving_currency_amount: int,
    receiving_currency: InvoiceCurrency,
    expiration: datetime,
    callback: str,
    is_subject_to_travel_rule: bool,
    signing_private_key: bytes,
    required_payer_data: Optional[CounterpartyDataOptions] = None,
    comment_chars_allowed: Optional[int] = None,
    receiver_kyc_status: Optional[KycStatus] = None,
    invoice_limit: Optional[int] = None,
    sender_uma: Optional[str] = None,
) -> Invoice:
    """
    Creates a UMA invoice.

    Args:
        receiver_uma: the UMA address of the receiver.
        receiving_currency_amount: the amount that the receiver will receive in either the smallest unit of the receiving currency.
        receiving_currency: the currency that the receiver will receive for this payment.
        expiration: the expiration time of the invoice in seconds.
        callback: the URL that the receiver will call to send UTXOs of the channel that the receiver used to receive the payment once it completes.
        is_subject_to_travel_rule: whether the receiver is a financial institution that requires travel rule information.
        signing_private_key: the private key of the VASP that is creating the invoice. This will be used to sign the request.
        required_payer_data: RequiredPayerData the data about the payer that the sending VASP must provide in order to send a payment.
        comment_chars_allowed: the number of characters that the receiver allows in the comment field.
        receiver_kyc_status: the KYC status of the receiver.
        invoice_limit: the maximum times of the invoice can be paid.
        sender_uma: the UMA address of the sender.
    """
    uuid = str(uuid4())
    invoice_required_payer_data = (
        InvoiceCounterpartyDataOptions(options=required_payer_data)
        if required_payer_data
        else None
    )
    invoice = Invoice(
        receiver_uma=receiver_uma,
        invoice_uuid=uuid,
        amount=receiving_currency_amount,
        receving_currency=receiving_currency,
        expiration=int(expiration.timestamp()),
        is_subject_to_travel_rule=is_subject_to_travel_rule,
        required_payer_data=invoice_required_payer_data,
        uma_versions=UMA_PROTOCOL_VERSION,
        comment_chars_allowed=comment_chars_allowed,
        sender_uma=sender_uma,
        max_num_payments=invoice_limit,
        kyc_status=receiver_kyc_status,
        callback=callback,
    )

    payload = invoice.to_tlv()
    signature = sign_payload(payload, signing_private_key)
    invoice.signature = bytes.fromhex(signature)
    return invoice


def verify_uma_invoice_signature(
    invoice: Invoice,
    other_vasp_pubkeys: PubkeyResponse,
) -> None:
    new_invoice = replace(invoice, signature=None)
    signature = invoice.signature
    if not signature:
        print("No signature in invoice")
    _verify_signature(
        new_invoice.to_tlv(),
        none_throws(signature).hex(),
        other_vasp_pubkeys.get_signing_pubkey(),
    )



# Generated error codes - DO NOT MODIFY MANUALLY

from enum import Enum
from dataclasses import dataclass


@dataclass
class ErrorDetails:
    code: str
    http_status_code: int


class ErrorCode(Enum):
    COUNTERPARTY_PUBKEY_FETCH_ERROR = ErrorDetails(
        code="COUNTERPARTY_PUBKEY_FETCH_ERROR", http_status_code=424
    )
    """Error fetching counterparty public key for validating signatures or encrypting messages"""

    INVALID_PUBKEY_FORMAT = ErrorDetails(
        code="INVALID_PUBKEY_FORMAT", http_status_code=400
    )
    """Error parsing the counterparty public key response"""

    CERT_CHAIN_INVALID = ErrorDetails(code="CERT_CHAIN_INVALID", http_status_code=400)
    """The provided certificate chain is invalid"""

    CERT_CHAIN_EXPIRED = ErrorDetails(code="CERT_CHAIN_EXPIRED", http_status_code=400)
    """The provided certificate chain has expired"""

    INVALID_SIGNATURE = ErrorDetails(code="INVALID_SIGNATURE", http_status_code=401)
    """The provided signature is not valid"""

    INVALID_TIMESTAMP = ErrorDetails(code="INVALID_TIMESTAMP", http_status_code=400)
    """The provided timestamp is not valid"""

    INVALID_NONCE = ErrorDetails(code="INVALID_NONCE", http_status_code=400)
    """The provided nonce is not valid"""

    INTERNAL_ERROR = ErrorDetails(code="INTERNAL_ERROR", http_status_code=500)
    """An unexpected error occurred on the server"""

    NON_UMA_LNURL_NOT_SUPPORTED = ErrorDetails(
        code="NON_UMA_LNURL_NOT_SUPPORTED", http_status_code=403
    )
    """This party does not support non-UMA LNURLs"""

    MISSING_REQUIRED_UMA_PARAMETERS = ErrorDetails(
        code="MISSING_REQUIRED_UMA_PARAMETERS", http_status_code=400
    )
    """Missing required UMA parameters"""

    UNSUPPORTED_UMA_VERSION = ErrorDetails(
        code="UNSUPPORTED_UMA_VERSION", http_status_code=412
    )
    """The counterparty UMA version is not supported"""

    PARSE_LNURLP_REQUEST_ERROR = ErrorDetails(
        code="PARSE_LNURLP_REQUEST_ERROR", http_status_code=400
    )
    """Error parsing the LNURLP request"""

    VELOCITY_LIMIT_EXCEEDED = ErrorDetails(
        code="VELOCITY_LIMIT_EXCEEDED", http_status_code=403
    )
    """This user has exceeded the velocity limit and is unable to receive payments at this time"""

    USER_NOT_FOUND = ErrorDetails(code="USER_NOT_FOUND", http_status_code=404)
    """The user for this UMA was not found"""

    USER_NOT_READY = ErrorDetails(code="USER_NOT_READY", http_status_code=403)
    """The user for this UMA is not ready to receive payments at this time"""

    REQUEST_NOT_FOUND = ErrorDetails(code="REQUEST_NOT_FOUND", http_status_code=404)
    """The request corresponding to this callback URL was not found"""

    PARSE_PAYREQ_REQUEST_ERROR = ErrorDetails(
        code="PARSE_PAYREQ_REQUEST_ERROR", http_status_code=400
    )
    """Error parsing the payreq request"""

    AMOUNT_OUT_OF_RANGE = ErrorDetails(code="AMOUNT_OUT_OF_RANGE", http_status_code=400)
    """The amount provided is not within the min/max range"""

    INVALID_CURRENCY = ErrorDetails(code="INVALID_CURRENCY", http_status_code=400)
    """The currency provided is not valid or supported"""

    SENDER_NOT_ACCEPTED = ErrorDetails(code="SENDER_NOT_ACCEPTED", http_status_code=400)
    """Payments from this sender are not accepted"""

    MISSING_MANDATORY_PAYER_DATA = ErrorDetails(
        code="MISSING_MANDATORY_PAYER_DATA", http_status_code=400
    )
    """Payer data is missing fields that are required by the receiver"""

    MISSING_MANDATORY_PAYEE_DATA = ErrorDetails(
        code="MISSING_MANDATORY_PAYEE_DATA", http_status_code=400
    )
    """Payee data is missing fields that are required by the sender"""

    UNRECOGNIZED_MANDATORY_PAYEE_DATA_KEY = ErrorDetails(
        code="UNRECOGNIZED_MANDATORY_PAYEE_DATA_KEY", http_status_code=501
    )
    """Receiver does not recognize the mandatory payee data key"""

    PARSE_UTXO_CALLBACK_ERROR = ErrorDetails(
        code="PARSE_UTXO_CALLBACK_ERROR", http_status_code=400
    )
    """Error parsing the utxo callback"""

    COUNTERPARTY_NOT_ALLOWED = ErrorDetails(
        code="COUNTERPARTY_NOT_ALLOWED", http_status_code=403
    )
    """This party does not accept payments with the counterparty"""

    PARSE_LNURLP_RESPONSE_ERROR = ErrorDetails(
        code="PARSE_LNURLP_RESPONSE_ERROR", http_status_code=400
    )
    """Error parsing the LNURLP response"""

    PARSE_PAYREQ_RESPONSE_ERROR = ErrorDetails(
        code="PARSE_PAYREQ_RESPONSE_ERROR", http_status_code=400
    )
    """Error parsing the payreq response"""

    LNURLP_REQUEST_FAILED = ErrorDetails(
        code="LNURLP_REQUEST_FAILED", http_status_code=424
    )
    """The LNURLP request failed"""

    PAYREQ_REQUEST_FAILED = ErrorDetails(
        code="PAYREQ_REQUEST_FAILED", http_status_code=424
    )
    """The payreq request failed"""

    NO_COMPATIBLE_UMA_VERSION = ErrorDetails(
        code="NO_COMPATIBLE_UMA_VERSION", http_status_code=424
    )
    """No compatible UMA protocol version found between sender and receiver"""

    INVALID_INVOICE = ErrorDetails(code="INVALID_INVOICE", http_status_code=400)
    """The provided invoice is invalid"""

    INVOICE_EXPIRED = ErrorDetails(code="INVOICE_EXPIRED", http_status_code=400)
    """The invoice has expired"""

    QUOTE_EXPIRED = ErrorDetails(code="QUOTE_EXPIRED", http_status_code=400)
    """The quote has expired"""

    INVALID_INPUT = ErrorDetails(code="INVALID_INPUT", http_status_code=400)
    """The provided input is invalid"""

    INVALID_REQUEST_FORMAT = ErrorDetails(
        code="INVALID_REQUEST_FORMAT", http_status_code=400
    )
    """The request format is invalid"""

    FORBIDDEN = ErrorDetails(code="FORBIDDEN", http_status_code=403)
    """This action is not permitted for this user"""

    NOT_IMPLEMENTED = ErrorDetails(code="NOT_IMPLEMENTED", http_status_code=501)
    """This functionality is not implemented"""

    QUOTE_NOT_FOUND = ErrorDetails(code="QUOTE_NOT_FOUND", http_status_code=404)
    """The requested quote was not found"""

    UMA_CONFIGURATION_FETCH_ERROR = ErrorDetails(
        code="UMA_CONFIGURATION_FETCH_ERROR", http_status_code=424
    )
    """Error fetching counterparty UMA configuration"""



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from dataclasses import dataclass
import json
from typing import Any, Dict

from uma.JSONable import JSONable
from uma.protocol.v0.currency import Currency as V0Currency
from uma.protocol.v1.currency import Currency as V1Currency
from uma.version import MAJOR_VERSION


@dataclass
class Currency(JSONable):
    code: str
    """
    ISO 4217 currency code (if applicable). For example, USD for US Dollars. For cryptocurrencies, this will
    be a ticker symbol, such as BTC for Bitcoin.
    """

    name: str
    """
    Full display name of the currency. For example, in USD, the name is "US Dollars".
    """

    symbol: str
    """
    Symbol for this currency. For example, in USD, the symbol is "$".
    """

    millisatoshi_per_unit: float
    """
    Estimated millisats per smallest "unit" of this currency (eg. 1 cent in USD).
    """

    min_sendable: int
    """
    Minimum amount that can be sent in this currency. This is in the smallest unit of the currency
    (eg. cents for USD).
    """

    max_sendable: int
    """
    Maximum amount that can be sent in this currency. This is in the smallest unit of the currency
    (eg. cents for USD).
    """

    decimals: int
    """
    The number of digits after the decimal point for display on the sender side, and to add clarity
	around what the "smallest unit" of the currency is. For example, in USD, by convention, there are 2 digits for
	cents - $5.95. In this case, `decimals` would be 2. Note that the multiplier is still always in the smallest
	unit (cents). In addition to display purposes, this field can be used to resolve ambiguity in what the multiplier
	means. For example, if the currency is "BTC" and the multiplier is 1000, really we're exchanging in SATs, so
	`decimals` would be 8.
	For details on edge cases and examples, see https://github.com/uma-universal-money-address/protocol/blob/main/umad-04-lnurlp-response.md.
    """

    uma_major_version: int = MAJOR_VERSION
    """
    The major version of the UMA protocol that this currency adheres to. This is not serialized to JSON.
    """

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {"millisatoshi_per_unit": "multiplier"}

    def to_dict(self) -> Dict[str, Any]:
        if self.uma_major_version == 0:
            return V0Currency(
                code=self.code,
                name=self.name,
                symbol=self.symbol,
                millisatoshi_per_unit=self.millisatoshi_per_unit,
                min_sendable=self.min_sendable,
                max_sendable=self.max_sendable,
                decimals=self.decimals,
            ).to_dict()
        return V1Currency(
            code=self.code,
            name=self.name,
            symbol=self.symbol,
            millisatoshi_per_unit=self.millisatoshi_per_unit,
            min_sendable=self.min_sendable,
            max_sendable=self.max_sendable,
            decimals=self.decimals,
        ).to_dict()

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        is_v0 = "minSendable" in json_dict
        # pylint: disable=protected-access
        currency_dict = (
            V0Currency._from_dict(json_dict)
            if is_v0
            else V1Currency._from_dict(json_dict)
        )
        currency_dict["uma_major_version"] = 0 if is_v0 else 1

        return currency_dict

    @classmethod
    def from_json(cls: "type[Currency]", json_encoded: str) -> "Currency":
        json_dict = json.loads(json_encoded)
        is_v0 = "minSendable" in json_dict
        if is_v0:
            v0_currency = V0Currency.from_json(json_encoded)
            return Currency(
                code=v0_currency.code,
                name=v0_currency.name,
                symbol=v0_currency.symbol,
                millisatoshi_per_unit=v0_currency.millisatoshi_per_unit,
                min_sendable=v0_currency.min_sendable,
                max_sendable=v0_currency.max_sendable,
                decimals=v0_currency.decimals,
                uma_major_version=0,
            )

        v1_currency = V1Currency.from_json(json_encoded)
        return Currency(
            code=v1_currency.code,
            name=v1_currency.name,
            symbol=v1_currency.symbol,
            millisatoshi_per_unit=v1_currency.millisatoshi_per_unit,
            min_sendable=v1_currency.min_sendable,
            max_sendable=v1_currency.max_sendable,
            decimals=v1_currency.decimals,
            uma_major_version=1,
        )



from typing import Optional, TypeVar


T = TypeVar("T")


def none_throws(value: Optional[T], error_message: Optional[str] = None) -> T:
    if value is None:
        raise RuntimeError(error_message or "Unexpected None")
    return value



from dataclasses import dataclass
from typing import Dict, List, Optional
from uma.JSONable import JSONable

from uma.exceptions import InvalidRequestException
from uma.generated.errors import ErrorCode
from uma.protocol.backing_signature import BackingSignature
from uma.protocol.counterparty_data import CounterpartyDataOptions
from uma.protocol.currency import Currency
from uma.protocol.kyc_status import KycStatus
from uma.protocol.settlement import SettlementOption
from uma.signing_utils import sign_payload


@dataclass
class LnurlComplianceResponse(JSONable):
    kyc_status: KycStatus
    """
    Whether the receiver is KYC verified by the receiving VASP.
    """

    signature: str
    """
    DER-encoded signature from the receiving VASP.
    """

    signature_nonce: str
    """
    A random string included in the signature payload to prevent replay attacks.
    """

    signature_timestamp: int
    """
    The time at which the request was made.
    """

    is_subject_to_travel_rule: bool
    """
    Whether the receiving VASP is subject to travel rule regulations.
    """

    receiver_identifier: str
    """
    The UMA address of the receiver.
    """

    backing_signatures: Optional[List[BackingSignature]] = None
    """
    List of backing VASP signatures.
    """


@dataclass
class LnurlpResponse(JSONable):
    tag: str
    callback: str
    """
    The URL that the sender will call for the payreq request.
    """

    min_sendable: int
    """
    The minimum amount that the sender can send in millisatoshis.
    """

    max_sendable: int
    """
    The maximum amount that the sender can send in millisatoshis.
    """

    encoded_metadata: str
    """
    JSON-encoded metadata that the sender can use to display information to the user.
    """

    currencies: Optional[List[Currency]]
    """
    The list of currencies that the receiver accepts in order of preference.
    """

    required_payer_data: Optional[CounterpartyDataOptions]
    """
    The data about the payer that the sending VASP must provide in order to send a payment.
    """

    compliance: Optional[LnurlComplianceResponse]
    """
    Compliance-related data from the receiving VASP.
    """

    uma_version: Optional[str]
    """
    The version of the UMA protocol that the receiver is using.
    """

    comment_chars_allowed: Optional[int] = None
    """
    The number of characters that the sender can include in the comment field of the pay request.
    """

    nostr_pubkey: Optional[str] = None
    """
    An optional nostr pubkey used for nostr zaps (NIP-57). If set, it should be a valid
    BIP-340 public key in hex format.
    """

    allows_nostr: Optional[bool] = None
    """
    Should be set to true if the receiving VASP allows nostr zaps (NIP-57).
    """

    settlement_options: Optional[List[SettlementOption]] = None
    """
    Optional list of settlement layers and assets supported by the receiver.
    If not specified, the payment will settle on Lightning using BTC as the settlement asset.
    """

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {
            "encoded_metadata": "metadata",
            "required_payer_data": "payerData",
            "comment_chars_allowed": "commentAllowed",
            "settlement_options": "settlementOptions",
        }

    def is_uma_response(self) -> bool:
        return self.uma_version is not None and self.compliance is not None

    def signable_payload(self) -> bytes:
        if not self.compliance:
            raise InvalidRequestException(
                "compliance field is required for signing or verifying.",
                ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
            )
        signable = "|".join(
            [
                self.compliance.receiver_identifier,
                self.compliance.signature_nonce,
                str(self.compliance.signature_timestamp),
            ]
        )
        return signable.encode("utf8")

    def append_backing_signature(self, signing_private_key: bytes, domain: str) -> None:
        """
        Appends a backing signature to the lnurlp response.

        Args:
            signing_private_key: The private key of the backing VASP which is used to sign the payload.
            domain: The domain of the backing VASP that produced the signature. Public keys for this VASP
            will be fetched from this domain at /.well-known/lnurlpubkey and used to verify the signature.
        """
        if not self.is_uma_response():
            return
        compliance = self.compliance
        if not compliance:
            raise InvalidRequestException(
                "compliance field is required for adding backing signatures.",
                ErrorCode.MISSING_REQUIRED_UMA_PARAMETERS,
            )
        payload = self.signable_payload()
        backing_signature = sign_payload(payload, signing_private_key)
        if compliance.backing_signatures is None:
            compliance.backing_signatures = []
        compliance.backing_signatures.append(
            BackingSignature(domain=domain, signature=backing_signature)
        )



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Dict
from uma.exceptions import InvalidNonceException


class INonceCache(ABC):
    @abstractmethod
    def check_and_save_nonce(self, nonce: str, timestamp: datetime) -> None:
        """
        Checks if the given nonce has been used before, and if not, saves it.
        If the nonce has been used before, or if timestamp is too old, raises InvalidNonceException.

        Args:
            nonce: the nonce to cache.
            timestamp: timestamp corresponding to the nonce.
        """

    @abstractmethod
    def purge_nonces_older_than(self, timestamp: datetime) -> None:
        """
        Purges all nonces older than the given timestamp. This allows the cache to be pruned.

        Args:
            timestamp: the timestamp before which nonces should be removed.
        """


class InMemoryNonceCache(INonceCache):
    """
    InMemoryNonceCache is an in-memory implementation of NonceCache.
    It is not recommended to use this in production, as it will not persist across restarts. You likely want to
    implement your own NonceCache that persists to a database of some sort.
    """

    def __init__(self, oldest_valid_timestamp: datetime) -> None:
        self._cache: Dict[str, datetime] = {}
        self._oldest_valid_timestamp = oldest_valid_timestamp

    def check_and_save_nonce(self, nonce: str, timestamp: datetime) -> None:
        if timestamp < self._oldest_valid_timestamp:
            raise InvalidNonceException("Timestamp is too old.")
        if nonce in self._cache:
            raise InvalidNonceException("Nonce has already been used.")
        self._cache[nonce] = timestamp

    def purge_nonces_older_than(self, timestamp: datetime) -> None:
        expired_nonces = [nonce for nonce, ts in self._cache.items() if ts < timestamp]
        for nonce in expired_nonces:
            del self._cache[nonce]
        self._oldest_valid_timestamp = timestamp



from coincurve.keys import PrivateKey


def sign_payload(payload: bytes, private_key: bytes) -> str:
    key = _load_private_key(private_key)
    signature = key.sign(payload)
    return signature.hex()


def _load_private_key(key: bytes) -> PrivateKey:
    try:
        return PrivateKey(key)
    except ValueError:
        return PrivateKey.from_der(key)



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from abc import ABC, abstractmethod
from typing import Optional

from uma.protocol.settlement import SettlementInfo


class IUmaInvoiceCreator(ABC):
    @abstractmethod
    def create_uma_invoice(
        self,
        amount_msats: int,
        metadata: str,
        receiver_identifier: Optional[str],
    ) -> str:
        pass

    def create_invoice_for_settlement_layer(
        self,
        amount_units: int,
        metadata: str,
        receiver_identifier: Optional[str],
        settlement_info: Optional[SettlementInfo],
    ) -> str:
        """
        Creates a payment request with settlement-agnostic parameters.
        Default implementation delegates to create_uma_invoice for backward compatibility.
        Implementations can override this for settlement-layer specific behavior.

        Args:
            amount_units: Amount in smallest units of the settlement asset.
            metadata: Metadata to include.
            receiver_identifier: Receiver's UMA address.
            settlement_info: Settlement info including the layer and asset chosen by the sender.

        Returns:
            Payment request string.
        """
        # pylint: disable=unused-argument
        return self.create_uma_invoice(amount_units, metadata, receiver_identifier)



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from uma.exceptions import *
from uma.generated.errors import ErrorCode, ErrorDetails
from uma.nonce_cache import InMemoryNonceCache, INonceCache, InvalidNonceException
from uma.protocol.backing_signature import (
    BackingSignature,
)
from uma.protocol.counterparty_data import (
    CounterpartyDataOption,
    CounterpartyDataOptions,
    CounterpartyDataKeys,
    create_counterparty_data_options,
)
from uma.protocol.currency import Currency
from uma.protocol.kyc_status import KycStatus
from uma.protocol.lnurlp_request import LnurlpRequest
from uma.protocol.settlement import SettlementAsset, SettlementInfo, SettlementOption
from uma.protocol.lnurlp_response import LnurlComplianceResponse, LnurlpResponse
from uma.protocol.invoice import (
    Invoice,
    InvoiceCounterpartyDataOptions,
    InvoiceCurrency,
)
from uma.protocol.payee_data import PayeeData
from uma.protocol.payer_data import (
    CompliancePayerData,
    PayerData,
    compliance_from_payer_data,
    create_payer_data,
)
from uma.protocol.payreq import PayRequest
from uma.protocol.payreq_response import (
    PayReqResponse,
    PayReqResponseCompliance,
    PayReqResponsePaymentInfo,
)
from uma.protocol.post_tx_callback import (
    PostTransactionCallback,
    TransactionStatus,
    UtxoWithAmount,
)
from uma.protocol.pubkey_response import PubkeyResponse
from uma.public_key_cache import (
    InMemoryPublicKeyCache,
    IPublicKeyCache,
    IAsyncPublicKeyCache,
)
from uma.signing_utils import sign_payload
from uma.type_utils import none_throws
from uma.uma import (
    create_compliance_payer_data,
    create_pay_req_response,
    create_pay_req_response_for_settlement_layer,
    create_pay_request,
    create_pay_request_with_payer_data,
    create_post_transaction_callback,
    create_pubkey_response,
    create_uma_invoice,
    create_uma_lnurlp_request_url,
    create_uma_lnurlp_response,
    fetch_public_key_for_vasp,
    gen_fetch_public_key_for_vasp,
    generate_nonce,
    get_vasp_domain_from_uma_address,
    is_uma_lnurlp_query,
    parse_lnurlp_request,
    parse_lnurlp_response,
    parse_pay_req_response,
    parse_pay_request,
    parse_post_transaction_callback,
    verify_pay_req_response_signature,
    verify_pay_req_response_backing_signatures,
    verify_pay_request_signature,
    verify_pay_request_backing_signatures,
    verify_post_transaction_callback_signature,
    verify_uma_invoice_signature,
    verify_uma_lnurlp_query_signature,
    verify_uma_lnurlp_query_backing_signatures,
    verify_uma_lnurlp_response_signature,
    verify_uma_lnurlp_response_backing_signatures,
)
from uma.uma_invoice_creator import IUmaInvoiceCreator
from uma.urls import is_domain_local
from uma.version import (
    UMA_PROTOCOL_VERSION,
    ParsedVersion,
    get_highest_supported_version_for_major_version,
    get_supported_major_versions,
    is_version_supported,
    select_highest_supported_version,
    select_lower_version,
)



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

import json
from typing import List
from uma.generated.errors import ErrorCode


class UmaException(Exception):
    def __init__(self, reason: str, error_code: ErrorCode) -> None:
        super().__init__(reason)
        self.reason = reason
        self.code = error_code.value.code
        self.http_status_code = error_code.value.http_status_code
        self.error_code = error_code

    def get_additional_params(self) -> dict:
        """Override this method in child classes to add additional parameters to the JSON output"""
        return {}

    def to_json(self) -> str:
        result = {
            "status": "ERROR",
            "reason": self.reason,
            "code": self.code,
            **self.get_additional_params(),
        }
        return json.dumps(result)

    @classmethod
    def from_json(cls: type["UmaException"], json_str: str) -> "UmaException":
        try:
            data = json.loads(json_str)
            error_code = ErrorCode[data["code"]]
            return cls(data["reason"], error_code)
        except (json.JSONDecodeError, KeyError):
            return cls(
                f"Failed to parse error JSON: {json_str}", ErrorCode.INTERNAL_ERROR
            )

    def to_http_status_code(self) -> int:
        return self.http_status_code

    def __reduce__(self):
        return (self.__class__, (self.reason, self.error_code))


class UnsupportedVersionException(UmaException):
    def __init__(
        self, unsupported_version: str, supported_major_versions: List[int]
    ) -> None:
        super().__init__(
            f"Version {unsupported_version} not supported.",
            ErrorCode.UNSUPPORTED_UMA_VERSION,
        )
        self.unsupported_version = unsupported_version
        self.supported_major_versions = supported_major_versions

    def get_additional_params(self) -> dict:
        return {"supportedMajorVersions": self.supported_major_versions}


class InvalidRequestException(UmaException):
    def __init__(
        self,
        reason: str = "Invalid request",
        error_code: ErrorCode = ErrorCode.INVALID_REQUEST_FORMAT,
    ):
        super().__init__(reason, error_code)


class InvalidCurrencyException(UmaException):
    def __init__(self, reason: str = "Invalid currency"):
        super().__init__(reason, ErrorCode.INVALID_CURRENCY)


class InvalidSignatureException(UmaException):
    def __init__(self, reason: str = "Cannot verify signature"):
        super().__init__(reason, ErrorCode.INVALID_SIGNATURE)


class InvalidNonceException(UmaException):
    def __init__(self, reason: str = "Invalid nonce"):
        super().__init__(reason, ErrorCode.INVALID_NONCE)



from dataclasses import dataclass

from uma.JSONable import JSONable


@dataclass
class BackingSignature(JSONable):
    domain: str
    """
    The domain of the backing VASP that produced the signature. Public keys for this VASP will be fetched
	from this domain at /.well-known/lnurlpubkey and used to verify the signature.
    """

    signature: str
    """
    Signature of the payload by a backing VASP that can attest to the authenticity of the message.
    """



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

import dataclasses
import json
from abc import ABC
from enum import Enum
from typing import Any, Dict, TypeVar

SELF = TypeVar("SELF", bound="JSONable")


@dataclasses.dataclass
class JSONable(ABC):
    def to_dict(self) -> Dict[str, Any]:
        json_dict = {}
        for key, value in self.__dict__.items():
            if isinstance(value, JSONable):
                value = value.to_dict()
            elif isinstance(value, list):
                value = [
                    item.to_dict() if isinstance(item, JSONable) else item
                    for item in value
                ]
            elif isinstance(value, dict):
                value = {
                    k: v.to_dict() if isinstance(v, JSONable) else v
                    for k, v in value.items()
                }
            elif isinstance(value, Enum):
                value = value.name
            if value is not None:
                json_dict[self._get_field_name(key)] = value
        return json_dict

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def _get_field_name_overrides(cls) -> Dict[str, str]:
        return {}

    @classmethod
    def _get_field_name(cls, name: str) -> str:
        name_overrides = cls._get_field_name_overrides()
        if name in name_overrides:
            return name_overrides[name]
        words = name.split("_")
        return words[0] + "".join(word.title() for word in words[1:])

    @classmethod
    def from_json(cls: "type[SELF]", json_encoded: str) -> SELF:
        data = cls._from_dict(json.loads(json_encoded))
        return cls(**data)

    @classmethod
    def _from_dict(cls, json_dict: Dict[str, Any]) -> Dict[str, Any]:
        # pylint: disable=protected-access
        data = {}
        for field in dataclasses.fields(cls):
            value = json_dict.get(cls._get_field_name(field.name), None)
            if value is None:
                data[field.name] = None
                continue

            field_type = field.type

            # handle Optional field
            if (
                hasattr(field.type, "__args__")
                and len(field.type.__args__) == 2
                and field.type.__args__[-1] is type(None)
            ):
                field_type = field.type.__args__[0]

            if isinstance(value, list) and hasattr(
                field_type.__args__[0], "_from_dict"
            ):
                # handle list
                field_type = field_type.__args__[0]
                data[field.name] = [
                    field_type(**field_type._from_dict(item)) for item in value
                ]
            elif hasattr(field_type, "_from_dict"):
                data[field.name] = field_type(**field_type._from_dict(value))
            elif isinstance(value, dict) and hasattr(
                field_type.__args__[1], "_from_dict"
            ):
                # handle dict
                field_type = field_type.__args__[1]
                data[field.name] = {
                    k: field_type(**field_type._from_dict(v)) for k, v in value.items()
                }
            elif isinstance(field_type, type) and issubclass(field_type, Enum):
                data[field.name] = field_type[value]
            else:
                data[field.name] = value

        return data



# Copyright ©, 2022-present, Lightspark Group, Inc. - All Rights Reserved

from enum import Enum


class KycStatus(Enum):
    VERIFIED = "VERIFIED"
    NOT_VERIFIED = "NOT_VERIFIED"
    PENDING = "PENDING"
    UNKNOWN = "UNKNOWN"

    def to_bytes(self) -> bytes:
        return self.value.encode()

    @classmethod
    def from_bytes(cls, data: bytes) -> "KycStatus":
        return cls(data.decode())



def is_domain_local(domain: str) -> bool:
    domain_tld = domain.split(".")[-1].split(":")[0]
    return (
        domain.startswith("localhost:")
        or domain.startswith("127.0.0.1:")
        or domain_tld == "local"
        or domain_tld == "internal"
    )
